'''
Created on Aug 9, 2022

@author: don_bacon
'''

from enum import Enum

class ActionType(Enum):
    """Action types referenced on Opportunity cards, occupation and border squares

    """
    EXTRA_TURN = 1          # player gets an extra turn
    LEAVE_UNEMPLOYMENT = 2  # player can leave unemployment without penalty
    COLLECT_EXPERIENCE = 3  # collect experience card from other players
    DRAW_EXPERIENCE = 4     # draw 1 or more experience card(s) from the deck
    DRAW_OPPORTUNITY = 5    # "opportunity"
    LOSE_TURN = 6            # "loseTurn"
    CASH_LOSS = 7            # "cashLoss"  percent or amount
    SALARY_INCREASE = 8        # "salaryIncrease" money amount
    SALARY_DIE_ROLL = 9        # "salary_dieRoll" amount x roll of 1 die
    BONUS_AMOUNT = 10        # "bonus_amount"  cash bonus amount
    SALARY_CUT = 11            # "salaryCut" by percent or amount
    BONUS_DIE_ROLL = 12        # cash amount x roll of 1 die "bonus_dieRoll"
    CALL_IN_FAVORS = 13        # "favors"
    CASH_LOSS_OR_UNEMPLOYMENT = 14        # cash loss amount or go to unemployment "cashLossOrUnemployment"
    EXIT_OCCUPATION = 15    # "occupationExit"
    ENTER_OCCUPATION = 16   # "enterOccupation"
    CASH_LOSS_DICE_ROLL = 17    # "cashLossRoll" salary, %salary, cash or %cash, 1 or 2 dice
    TRAVEL = 18                # "travel_shortcut" or "travel_border" next_square = number or name
    OCCUPATION_EXIT = 19    # "occupationExit"  raise salary if College or give Experience cards
    PAY_TAX = 20            # "payTax" as %salary depending on salary    
    
    def describe(self):
        # self is the member here
        return self.name, self.value
    


'''
Created on Aug 9, 2022

@author: don_bacon
'''

from game.careersObject import CareersObject

class OpportunityCard(CareersObject):
    """
    Represents a single Opportunity card
    """

    def __init__(self, opportunity_type, number, ncard, destination=None, text="", expenses_paid=False, double_happiness=False):
        """Constructor
            Arguments:
                opportunity_type - the OpportunityType of this card
                number - the unique number identifying this card
                ncard - the ordinal of this type of card (from 0 to quantity-1)
                destination - the "name" of the corresponding border square if there is a destination, otherwise None
                text - the Opportunity card text
                expenses_paid - Applies to OpportunityType.OCCUPATION: True if the player can enter for free.
                double_happiness - Applies to OpportunityType.OCCUPATION: True if happiness point values are doubled.
        """
        self._opportunity_type = opportunity_type
        if destination == "":
            self._destination = None
        else:
            self._destination = destination
        self._text = text
        self._expenses_paid = expenses_paid
        self._double_happiness = double_happiness
        self._action_type = None
        self._ncard = ncard
        self._number = number
        
    @property
    def opportunity_type(self) ->str:
        return self._opportunity_type
    
    @opportunity_type.setter
    def opportunity_type(self, ctype:str):
        self._opportunity_type = ctype
    
    @property
    def destination(self):
        return self._destination
    
    @destination.setter
    def destination(self, bs):
        self._destination = bs
    
    @property
    def text(self):
        return self._text
    
    @text.setter
    def text(self, text):
        self._text = text
    
    @property
    def expenses_paid(self):
        return self._expenses_paid
    
    @expenses_paid.setter
    def expenses_paid(self, ep):
        self._expenses_paid = ep
    
    @property
    def double_happiness(self):
        return self._double_happiness
    
    @double_happiness.setter
    def double_happiness(self, dh):
        self._double_happiness = dh

    @property
    def action_type(self):
        return self._action_type
    
    @action_type.setter
    def action_type(self, value):
        self._action_type = value
        
    @property
    def ncard(self):
        return self._ncard
    
    @property
    def number(self):
        return self._number

    def __str__(self):
        return self.text
    
    def to_JSON(self):
        jtxt = f'{{"number" : "{self.number}", '
        jtxt += f'"text" : "{self.text}" }}'
        return jtxt
    
    


'''
Created on Aug 12, 2022

@author: don_bacon
'''

from game.player import  Player
from game.careersGameEngine import CareersGameEngine
from game.commandResult import CommandResult

class GameRunner(object):
    """A command-line text version of Careers game play used for testing and simulating web server operation.
    """

    def __init__(self, edition, master_id, game_type, total_points):
        """
        Constructor
        """
        self._careersGame = None
        self.total_points = total_points
        self._trace = True          # traces the action by describing each step
        self._edition = edition
        self._game_type = game_type
        self.game_engine = CareersGameEngine()
        self.game_engine.trace = self._trace
        self._master_id = master_id

    def add_player(self, name, initials, stars=0, hearts=0, cash=0):
        self.game_engine.add(name, initials, stars, hearts, cash)
    
    @property
    def trace(self):
        return self._trace
    
    @trace.setter
    def trace(self, value):
        self._trace = value
        
    @property
    def careersGame(self):
        return self._careersGame
    
    @property
    def master_id(self):
        return self._master_id
    
    
    def number_of_players(self):
        return self.game.game_state.number_of_players
    
    def get_game_state(self):
        return self.game_engine.game_state
    
    def execute_command(self, cmd, aplayer:Player):
        """Command format is command name + optional arguments
            The command and arguments passed to the game engine for execution
            Arguments:
                cmd - the command string
                aplayer - a Player reference. If None, the engine executes the command as Administrator (admin) user.
            Returns:
                result - result dictionary (result, message, done)
        """

        args = []
            
        result = self.game_engine.execute_command(cmd, aplayer, args)
        return result
    
    def run_game(self):

        game_over = False
        game_state = self.get_game_state()
        nplayers = game_state.number_of_players
        turn_number = 1
        while not game_over:
            for i in range(nplayers):
                # 
                current_player = game_state.current_player
                pn = current_player.number
                done = False
                while not done:
                    prompt = f'player {pn} {current_player.player_initials} ({turn_number}): '
                    cmd = input(prompt)
                    result = self.execute_command(cmd, current_player)
                    print(result.message)
                    done = result.done_flag
                    if result.return_code == CommandResult.TERMINATE:
                        game_over = True
                if game_over:
                    break
            turn_number += 1
        
        self.game_engine.end()
        

if __name__ == '__main__':

    total_points = 100
    edition = 'Hi-Tech'
    game_type = 'points'            # or 'timed'
    master_id = 'ZenAlien2013'      # uniquely identifies 'me' as the game creator
    game_runner = GameRunner(edition, master_id, game_type, total_points)  # creates a CareersGameEngine
    game_runner.execute_command(f'create {edition} {master_id} {game_type} {total_points}', None)     # creates a CareersGame for points
    #
    # add 2 players
    #
    game_runner.execute_command("add player Don DWB 40 10 50", None)
    game_runner.execute_command("add player Brian BDB 20 40 40", None)
    game_runner.execute_command("start", None)
    
    game_runner.run_game()
    
    



'''
Created on Aug 14, 2022

@author: don_bacon
'''
from game.gameSquare import GameSquare
from game.player import Player
from game.commandResult import CommandResult

class OccupationSquare(GameSquare):
    '''
    classdocs
    '''


    def __init__(self, occupation_square_dict, game=None):
        """Create a OccupationSquare instance.
            Arguments:
                occupation_square_dict - the dictionary defining this OccupationSquare. This would be an element of occupationSquares.
                game - a CareersGame instance

        """
        super().__init__("Occupation", name=None, number= occupation_square_dict['number'], \
                         text=occupation_square_dict['text'], special_processing_dict=occupation_square_dict['specialProcessing'], game=game)
        
        self._game_square_dict = occupation_square_dict
        self._stars = occupation_square_dict["stars"]
        self._hearts = occupation_square_dict["hearts"]
        self._experience = occupation_square_dict["experience"]         # the number of Experience cards to collect on this square
        self._opportunities = occupation_square_dict["opportunities"]   # the number of Opportunity cards to collect on this square
        
        self.action_text = occupation_square_dict.get('action_text', None)
        self._bonus = occupation_square_dict.get('bonus',0)
        
        
    @property
    def stars(self):
        return self._stars
    
    @property
    def hearts(self):
        return self._hearts
    
    @property
    def experience(self):
        return self._experience
    
    @property
    def opportunities(self):
        return self._opportunities
    
    def execute(self, player:Player) -> CommandResult:
        """Executes the actions associated with this occupation square for a given Player
            Returns: CommandResult
        """
        result = CommandResult(0, "TODO", False)   #  TODO
        return result
    
    
    def to_JSON(self):
        txt = f'''{{
        "square_class" : "{self.square_class}",
        "name":{self.name},
        "number":"{self.number}",
        "text":"{self.text}",
        "stars":"{self.stars}",
        "hearts":"{self.hearts}",
        "experience":"{self.experience}",
        "opportunities":"{self.opportunities}",
        "action_text":"{self.action_text}",
        "special_processing_txt" : {self._special_processing_dict} 
        }}'''
        return txt
        
    


'''
Created on Aug 9, 2022

@author: don_bacon
'''

from game.cardDeck import CardDeck
from game.experienceCard import ExperienceCard

class ExperienceCardDeck(CardDeck):
    """The deck of Experience Cards used in game play.
    
    """

    def __init__(self, resource_path, edition_name):
        '''
        Constructor
        '''
        super().__init__(resource_path, "experienceCards", edition_name)    # loads the card deck
        
    def save_card(self, card_spec, qty):
        """Appends n-instances of a single Experience card to the card deck.
            Arguments: card_spec - the JSON experience card content as a dict
            
        card_spec includes "quantity" tag that specifies the number of this cards that appear in the deck.

        """
        number = card_spec['number']
        for ncard in range(1, qty+1):
            experience_card = ExperienceCard(number, ncard, card_spec['type'], card_spec['spaces'])
            self._deck.append(experience_card)
    


from __future__ import absolute_import  # multi-line and relative/absolute imports

from .__version__ import __title__, __description__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__


__all__ = [
    'actionType',
    'boardLocation',
    'borderSquare',
    'cardDeck',
    'careersGame',
    'careersGameEngine',
    'careersObject',
    'commandResult',
    'experienceCard',
    'experienceCardDeck',
    'experienceCardType',
    'gameBoard',
    'gamePlayer',
    'gameSquare',
    'gameState',
    'gameUtils',
    'occupation',
    'occupationSquare',
    'opportunityCard',
    'opportunityCardDeck',
    'opportunityType',
    'player',
    'specialProcessing',
    'successFormula'
]

from careers.game.actionType import ActionType
from careers.game.boardLocation import BoardLocation
from careers.game.borderSquare import BorderSquare
from careers.game.gameSquare import GameSquare
from careers.game.gameState import GameState
from careers.game.gameBoard import GameBoard
from careers.game.occupation import Occupation
from careers.game.occupationSquare import OccupationSquare
from careers.game.careersObject import CareersObject
from careers.game.cardDeck import CardDeck
from careers.game.careersGame import CareersGame
from careers.game.careersGameEngine import CareersGameEngine
from careers.game.gameRunner import GameRunner
from careers.game.gameUtils import GameUtils
from careers.game.player import Player
from careers.game.successFormula import SuccessFormula
from careers.game.opportunityType import OpportunityType
from careers.game.opportunityCard import OpportunityCard
from careers.game.opportunityCardDeck import OpportunityCardDeck
from careers.game.experienceCard import ExperienceCard
from careers.game.experienceCardType import ExperienceCardType
from careers.game.experienceCardDeck import ExperienceCardDeck
from careers.game.specialProcessing import SpecialProcessing
from careers.game.commandResult import CommandResult




'''
Created on Aug 9, 2022

@author: don_bacon
'''

from game.careersObject import CareersObject

class ExperienceCard(CareersObject):
    """
    Represents a single Experience card
    """
    
    card_types = [ "fixed", "1-die wild", "2-dice wild", "triple wild" ]

    def __init__(self, number, ncard, experience_card_type, spaces):
        """
        Constructor
            Arguments:
                number - the unique number identifying this card
                experience_card_type - the type of this card, one of card_types: "fixed", "1-die wild", "2-dice wild", "triple wild" 
                spaces - the number of spaces to move, can be negative to move backwards.
                         spaces == 0 indicates a "wild" card as identified by experience_card_type
                ncard - the ordinal of this type of card (from 0 to quantity-1)
        """
        self._number = number
        self._card_type = experience_card_type
        self._spaces = spaces       # can be negative
        self._ncard = ncard
        
    @property
    def card_type(self) -> str:
        return self._card_type
    
    @card_type.setter
    def card_type(self, ct:str) :
        self._card_type = ct
        
    @property
    def spaces(self) -> int:
        return self._spaces
    
    @spaces.setter
    def spaces(self, spaces:int):
        self._spaces = spaces
    
    @property
    def number(self):
        return self._number

    @property
    def ncard(self):
        return self._ncard

    def __str__(self):
        if self.card_type == 'fixed':
            return f'{self.spaces} spaces'
        else:
            return self.card_type

    def to_JSON(self):
        jtxt = f'{{"number" : "{self.number}", '
        if self.card_type == 'fixed':
            jtxt += f'"spaces" : "{self.spaces}" }}'
        elif self.card_type == '1-die wild':
            jtxt += f'"spaces" : "?" }}'
        elif self.card_type == '2-die wild':
            jtxt += f'"spaces" : "??" }}'
        elif self.card_type == '3-die wild':
            jtxt += f'"spaces" : "???" }}'
        
        return jtxt
    
    


'''
Created on Aug 6, 2022

@author: don_bacon
'''

from careers.environment import Environment
from game.player import Player
import json, random
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional, Union, List

from game.opportunityCardDeck import OpportunityCardDeck
from game.experienceCardDeck import ExperienceCardDeck
from game.gameState import GameState
from game.occupation import Occupation
from game.gameBoard import GameBoard
from game.borderSquare import BorderSquare
from game.careersObject import CareersObject
from threading import Lock

class CareersGame(CareersObject):
    """
    Represents a Careers Game instance.
        The CareersGame model class has instances of all the components used in game play:
        Opportunity and Experience card decks, the game board, parameters for a given edition, occupations, players, and the game state.
        
    """
    _lock = Lock()
    
    def __init__(self, edition_name,  master_id, total_points, game_id, game_type="points"):
        """CareersGame Constructor
            Arguments:
                master_id - An ID that uniquely identifies the game's creator - a.k.a the game master
                edition_name - the name of the edition to create. This must be a key in editions.json file.
                total_points - total number of points in the success formula or in a timed game, the number of minutes.
                game_id - a Globally Unique Identifier for this game (guid). If not provided, one is generated from the current date/time.
            Raises:
                ValueError if there is no such edition
            Saved games are indexed by master_id. This is the primary search key used to search for saved games.
            It's the responsibility of the front end GUI to provide this.
        """
        self._master_id = master_id
        self._edition_name = edition_name
        self._env = Environment.get_environment()
        self._resource_folder = self._env.get_resource_folder()     # base resource folder
        #
        # validate the Edition
        #
        fp = open(self._resource_folder + "/editions.json", "r")
        jtxt = fp.read()
        editions_txt = json.loads(jtxt)
        self._editions = editions_txt['editions']   # list of editions
        if edition_name in self._editions:
            self._edition = editions_txt[edition_name]
        else: raise ValueError(f"No such edition {edition_name}")
        fp.close()
        #
        # load game parameters, layout and occupations
        #
        self._load_game_configuration()
 
        #
        # create the Opportunity and Experience card decks
        #
        self._create_opportunity_deck()
        self._create_experience_deck()
        self._college_degrees =  self._load_college_degrees()  # dict with keys: degreePrograms (list), maxDegrees (int), degreeNames (list)
        #
        # load the game board
        #
        self._game_board = self._create_game_board()   # GameBoard instance
        self._game_type = game_type  # 'points' or 'timed' (which is not yet supported)

        #
        # create & initialize the GameState which includes a list of Players
        #
        self._game_state = GameState(total_points, self._game_type)
        #
        # create a unique ID for this game, used for logging
        #
        self._gameId = game_id
        #
        # if a gameId is not provided, create one
        #
        if game_id is None:
            self._gameId = self._create_game_id()
        
    def _load_game_configuration(self):
        """Loads the game parameters and occupations JSON files for this edition.
        
        """
        fp = open(self._resource_folder + "/gameParameters_" + self._edition_name + ".json", "r")
        jtxt = fp.read()
        self._game_parameters = json.loads(jtxt)
        fp.close()

        fp = open(self._resource_folder + "/occupations_" + self._edition_name + ".json", "r")
        jtxt = fp.read()
        occupations_dict = json.loads(jtxt)
        self._occupation_names = occupations_dict['occupations']
        fp.close()
        
        # load the individual occupation files
        self._occupations = self.load_occupations()     # dictionary of Occupation instances keyed by name

    def _create_game_id(self) ->str :
        """Create a unique game id (guid) for this game.
            This method acquires a lock to insure uniqueness in a multi-process/web environment.
            Format is based on current date and time, for example 20220908-140952-973406-27191
        """
        CareersGame._lock.acquire()
        today = datetime.now()
        gid =  '{0:d}{1:02d}{2:02d}-{3:02d}{4:02d}{5:02d}-{6:06d}-{7:05d}'.format(today.year, today.month, today.day, today.hour, today.minute, today.second, today.microsecond, random.randint(10000,99999))
        CareersGame._lock.release()
        return gid
    
    def _create_opportunity_deck(self):
        self._opportunities = OpportunityCardDeck(self._resource_folder, self._edition_name)
    
    def _create_experience_deck(self):
        self._experience_cards = ExperienceCardDeck(self._resource_folder, self._edition_name)
        
    def _create_game_board(self) -> GameBoard:
        game_layout_filename = self._resource_folder + "/gameLayout_" + self._edition_name + ".json"
        game_board = GameBoard(game_layout_filename, game=self)
        return game_board

    def load_occupations(self) -> Dict[str, Occupation]:
        """Loads individual occupation JSON files for this edition.
            Arguments: occupation_list - a list of occupation names
            Returns: a dict with the occupation name as the key and contents
                of the corresponding occupation JSON file (as a dict) as the value.
                If the occupation JSON file doesn't exist, the value is None.
            Note: the filename is the occupation name + "_" + the edition_name + ".json"
                The file path is the resource_folder set in the Environment
        """
        occupations:Dict[str, Occupation] = {}
        for name in self._occupation_names:
            filepath = self._resource_folder + "/" + name + "_" + self._edition_name + ".json"
            p = Path(filepath)
            if p.exists():
                fp = open(filepath, "r")
                occupation_dict = json.loads(fp.read())
                # create an Occupation object for this occupation
                occupation = Occupation(occupation_dict)
                occupations[name] = occupation
            else:
                occupations[name] = None
        return occupations

    
    def _load_college_degrees(self) -> Dict[str, Union[List[int] ,List[str], int]]:
        fp = open(self._resource_folder + "/collegeDegrees_" + self._edition_name + ".json", "r")
        degrees = json.loads(fp.read())
        return degrees
    
    @property
    def edition(self):
        return self._edition
    
    @property
    def edition_name(self):
        return self._edition_name
    
    @property
    def game_board(self) -> GameBoard:
        return self._game_board
    
    def get_border_square(self, num:int) -> BorderSquare:
        """Convenience method to get a BorderSquare instance from the GameBoard
        """
        return self._game_board.get_square(num)
    
    def get_game_layout_dimensions(self) -> dict:
        """size: (number of squares), sides: (4-element list of #squares/side)
        """
        return self.game_board.game_layout_dimensions
    

    def get_game_board_size(self) -> int:
        return self.game_board.game_board_size
    
    
    @property
    def game_parameters(self):
        return self._game_parameters
    
    @property
    def game_state(self):
        return self._game_state
    
    @property
    def gameId(self):
        return self._gameId
    
    @property
    def master_id(self):
        return self._master_id
    
    @master_id.setter
    def master_id(self, value):
        self._master_id = value
    
    @property
    def game_type(self):
        return self._game_type
    
    @property
    def occupation_names(self) ->list:
        """List of occupation names (keys) for this edition
        """
        return self._occupation_names
    
    @property
    def occupations(self) -> dict:
        """Dictionary of Occupation instances keyed by occupation name
        """
        return self._occupations
    

    def get_occupation_entrance_squares(self) -> dict:
        return self.game_board.occupation_entrance_squares
    
    def get_travel_squares(self) -> list:
        return self.game_board.travel_squares
    
    def get_corner_squares(self) -> dict:
        return self.game_board.corner_squares
    
    @property
    def opportunities(self):
        return self._opportunities
    
    @property
    def experience_cards(self):
        return self._experience_cards
    
    @property
    def college_degrees(self):
        return self._college_degrees
   
    def add_player(self, aplayer:Player):
        self.game_state.add_player(aplayer)
    
    def complete_player_move(self) -> Optional[Player]:
        """Completes the move of the current player and determines if there's a winner and returns winning_player.
            If so winning_player is set. Otherise, current_player and current_player_number advanced to the next player.
            Returns: winning_player (Player instance) or None
        
        """
        winner = None
        if self.is_game_complete():       # sets winning_player if True
            winner = self.winning_player
        else:
            self.next_player_number()
        return winner
            
    
    def is_game_complete(self) -> bool:
        """Iterates over the players to see who, if anyone, has won.
            Returns: True if there's a winner, else False.
                    sets self.winning_player to the Player who won.
            NOTE - returns the first winning player if there happens to be more than 1.
            To prevent this from happening, this should be called at the end of each player's turn.
        """
        completed = False
        for p in self.players:
            if p.is_complete():
                self._winning_player = p
                completed = True
                break
        return completed
    
    def start_game(self):
        """
        Initializes game state and starts the game
        TODO
        """
        pass
    
    def complete_turn(self):
        """Completes the turn of the current_player
        
        """
        pass
    
    def find_next_border_square(self, current_square_number, atype:str, name:str=None) -> int:
        """Find the next border square of a given type and optional name
            Arguments:
                current_square_number - square number on the game board
                game_board - a GameBoard instance
                atype - the square type to look for. Valid values are 'travel_square', 'corner_square', 'travel_square', and 'occupation_entrance_square'
                name - optional square.name to use for comparison. For example in the UK edition, the travel squares
                       all have different names.
            Returns: the next square number OR None if no square of a given type/name is found.
            BTW, the next_square_number could be < current_square_number if it's past Payday
        """
        squares = []
        next_square_num = None
        if atype=='travel_square':
            squares = self.game_board.travel_squares
        elif atype=='opportunity_square':
            squares=self.game_board.opportunity_squares
        elif atype=='occupation_entrance_square':
            squares = self.game_board.occupation_entrance_squares
        elif atype=='corner_square':
            squares = self.game_board.corner_squares
        
        for square in squares:      # square is a BorderSquare instance, they are in square# order
            if square.square_type == atype and (name is None or square.name==name):
                if square.number > current_square_number:
                    next_square_num = square.number
                    break
        if next_square_num is None:     # wrap-around the board
            next_square_num = squares[0].number
        return next_square_num
    
    def find_border_square(self, name:str) -> Optional[BorderSquare]:
        bs = None
        for square in self.game_board.border_squares:
            if square.name == name:
                bs = square
                break
        return bs
    
    def to_JSON(self) -> str:
        """Use jsonpickle to serialize the game to JSON as a JSON-compatible dict.
            A CareersGame serialized with jsonpickle can be turned back to python with unpickler.
        """
        return CareersObject.json_pickle(self)
    
if __name__ == '__main__':
    game = CareersGame('Hi-Tech')
    print(game.occupation_list)
    
    



'''
Created on Aug 9, 2022

@author: don_bacon
'''
from enum import Enum

class OpportunityType(Enum):
    """Opportunity card types
    
    """
    OCCUPATION = 1
    OCCUPATION_CHOICE = 2
    BORDER = 3
    BORDER_CHOICE = 4
    ACTION = 5      # perform a non-movement action, qualified by ActionType
    TRAVEL = 6
    OPPORTUNITY = 7
        
    def describe(self):
        # self is the member here
        return self.name, self.value
    


'''
Created on Aug 12, 2022

@author: don_bacon
'''
import random
from datetime import datetime
from game.successFormula import SuccessFormula


class GameUtils(object):
    """
    General purpose routines.
    """

    def __init__(self, params):
        pass
    
    @staticmethod
    def shuffle(size:int) -> list:
        return random.sample(list(range(0, size)), size)
    
    
    @staticmethod
    def get_random_formula(total_points) -> SuccessFormula:
        cash = random.randint(1, total_points)
        hearts = random.randint(1, 100-cash)
        stars = random.randint(1,100-(cash+hearts))
        return SuccessFormula(stars, hearts, cash)
    
    @staticmethod
    def roll(number_of_dice):
        return random.choices(population=[1,2,3,4,5,6],k=number_of_dice)
    
    @staticmethod
    def get_datetime() -> str:
        """Returns - now formatted as a string, for example: 20220911_120545
        """
        now = datetime.today()
        return '{0:d}{1:02d}{2:02d}_{3:02d}{4:02d}{5:02d}'.format(now.year, now.month, now.day, now.hour, now.minute, now.second)
    
    

        


'''
Created on Aug 26, 2022

@author: don_bacon
'''

import json
from game.borderSquare import BorderSquare


class GameBoard(object):
    """Encapsulates a Careers game board border squares.
        This structure does not include the Occupation path squares.
        They are accessed from CareersGame from occupations dictionary.
    """

    def __init__(self, game_layout_filename, game=None):
        """Create a new GameBoard
            Arguments:
                game_layout_filename - the full path to the gameLayout JSON file, for example: "/Compile/careers/resources/gameLayout_Hi-Tech.json"
                game - a CareersGame instance
        """
        self._game_layout_filename = game_layout_filename
        self._occupation_entrance_squares = {}         # dictionary of BorderSquare that are type "occupation_entrance_square" indexed by name
        self._travel_squares = []                      # list of BorderSquare that are type "travel_square"
        self._corner_squares = {}                      # dict of BorderSquare that are type "corner_square" indexed by name
        self._opportunity_squares = []                 # list of BorderSquare that are type 'opportunity_square'
        self._action_squares = {}                      # dict of BorderSquare that are type "action_square" indexed by name
        self._danger_squares = {}                      # dict of BorderSquare that are type "danger_square" indexed by name
        fp = open(game_layout_filename, "r")
        self._game_board_dict = json.loads(fp.read())
        fp.close()
        self._game_layout = self._game_board_dict['layout']
        self._game_layout_dimensions = self._game_board_dict['dimensions']
        self._game_board_size = self._game_layout_dimensions['size']
        self._types = self._game_board_dict['type_list']        # a list of border square types
        
        self._border_squares = list()
        for border_square_dict in self._game_layout:
            border_square = BorderSquare(border_square_dict, game=game)
            #
            # populate square type lists/dictionaries
            #
            self._border_squares.append(border_square)
            
            if border_square.square_type == "occupation_entrance_square":
                self._occupation_entrance_squares[border_square.name] = border_square
                
            if border_square.square_type == "travel_square":
                self._travel_squares.append(border_square)
        
            if border_square.square_type == "corner_square":
                self._corner_squares[border_square.name] = border_square
                
            if border_square.square_type == "opportunity_square":
                self._opportunity_squares.append(border_square)
            
            if border_square.square_type == "action_square":
                self._action_squares[border_square.name] = border_square
            
            if border_square.square_type == "danger_square":
                self._danger_squares[border_square.name] = border_square
                
            if 'special_processing' in border_square_dict:
                pass
    
    @property
    def border_squares(self) ->list:    # list of BorderSquare
        return self._border_squares
    
    def get_square(self, num) -> BorderSquare:
        return self._border_squares[num]
    
    @property
    def game_layout(self) ->list:       # list of dict
        return self._game_layout
    
    @property
    def game_layout_dimensions(self) ->dict:
        return self._game_layout_dimensions

    @property
    def game_board_size(self):
        return self._game_board_size
    
    @property
    def occupation_entrance_squares(self) ->dict:
        return self._occupation_entrance_squares
    
    @property
    def travel_squares(self) ->list:
        return self._travel_squares
    
    @property
    def corner_squares(self) ->dict:
        return self._corner_squares
    
    @property
    def opportunity_squares(self):
        return self._opportunity_squares
    
    @property
    def action_squares(self):
        return self._action_squares
    
    @property
    def danger_squares(self):
        return self._danger_squares
    
    @property
    def types(self) ->list:
        return self._types
    
    


'''
Created on Aug 5, 2022

@author: don_bacon
'''

from game.careersObject import CareersObject
import json

class SuccessFormula(CareersObject):
    '''
    Encapsulates a player's success formula for a given game.
    '''

    def __init__(self, stars=0, hearts=0, cash=0):
        """Initialize the success formula with the number of stars (Fame), hearts (Happiness) and cash (money in 1000s) specified.
            Computes game_total_points as the sum.
        
        """
        self._fame = stars
        self._happiness = hearts
        self._cash = cash
        self._game_total_points = stars + hearts + cash
        
    @property
    def fame(self):
        return self._fame
    @property
    def happiness(self):
        return self._happiness
    @property
    def cash(self):
        return self._cash
    @fame.setter
    def fame(self, value):
        self._fame = value
    @happiness.setter
    def hapiness(self, value):
        self._hapiness = value
    @cash.setter
    def cash(self, value):
        self._cash = value
    
    def __str__(self):
        return f'Money: ${self.cash},000  Fame: {self.fame}  Happiness: {self.happiness}'
    
    def to_dict(self):
        return {"cash" : self.cash, "fame" : self.fame, "happiness" : self.happiness}

    def to_JSON(self):
        """Returns the JSON serialization of SuccessFormula.
        """
        return json.dumps(self.to_dict())
    
    @property
    def total_points(self):
        return self._game_total_points
    
    


'''
Created on Aug 14, 2022

@author: don_bacon
'''

from game.specialProcessing import SpecialProcessing
from game.careersObject import CareersObject
from game.player import Player
from game.commandResult import CommandResult
from typing import Dict, Union

class GameSquare(CareersObject):
    """Represents a square a player can land on. This can be a Border or Occupation.
    
    """
    
    GAME_SQUARE = Dict[str, Union[int, str, SpecialProcessing.SPECIAL_PROCESSING ]]

    def __init__(self, square_class:str, name=None, number=-1, text=None, special_processing_dict=None, action_text=None, game=None):
        """Create a GameSquare
        
        """
        self._square_class = square_class
        self._name = name
        self._number = number
        self._text = text
        self._action_text = action_text
        self._special_processing_dict = special_processing_dict
        self._game_square_dict = None       # populated by concrete class
    
        self._special_processing = SpecialProcessing(special_processing_dict, square_class) if special_processing_dict is not None and len(special_processing_dict) > 0 else None
        self._careersGame = None
        #
        # avoids circular import
        #
        if game is not None:
            from game.careersGame import CareersGame
            assert isinstance(game,CareersGame)
            self._careersGame = game
        
    @property
    def square_class(self):
        """Border or Occupation
        """
        return self._square_class
    
    @square_class.setter
    def square_class(self, value):
        if value is not None and ( value == 'Border' or value == 'Occupation'):
            self._square_class = value
        else: raise(ValueError("Invalid square class: " + value))
    
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, value):
        self._name = value
        
    @property
    def number(self):
        return self._number
    
    @number.setter
    def number(self, value):
        self._number = value
        
    @property
    def text(self):
        return self._text
    
    @text.setter
    def text(self, value):
        self._text = value
    
    @property
    def action_text(self):
        return self._action_text
    
    @action_text.setter
    def action_text(self, value):
        self._action_text = value
        
    @property
    def special_processing(self):
        return self._special_processing
    
    def execute(self, player:Player) -> CommandResult:
        """Execute actions associated with this Occupation or Border square
            Override in derived class. Base implementation returns None.
        
        """
        return None
    
    def execute_special_processing(self, player):
        """Invokes the specialProcessing section of this game square for a player.
            This method is called as exit processing for a square, whereas execute() is called when a player lands on a square.
            Override in derived class. Base implementation returns None.
        """
        return None
        
    def to_JSON(self):
        #
        # override in derived class
        #
        return None
    


from __future__ import absolute_import  # multi-line and relative/absolute imports

from .__version__ import __title__, __description__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__


__all__ = [

]



'''
Created on Aug 12, 2022

@author: don_bacon
'''

from game.careersObject import CareersObject
import json
from typing import Dict, List, Union

class SpecialProcessing(CareersObject):
    """Encapsulates the "specialProcessing" section of a border square or occupation square.
        This is only a model class. The actual processing is done by CareersGameEngine.
    """
    
    border_types = ["payday", "opportunity", "payTax", "enterOccupation", "enterCollege", "buyHearts", "buyStars", "buyExperience",\
                    "hospital", "unemployment", "buyInsurance", "gamble" ]

    occupation_types = ["loseNextTurn", "shortcut", "cashLossOrUnemployment", "goto",\
                         "salaryIncrease", "salaryCut", "bonus", "favors", "backstab", "fameLoss", "hapinessLoss"]
    
    common_types = ["travelBorder", "cashLoss", "extraTurn"]
    
    all_types = border_types + occupation_types + common_types
    
    SPECIAL_PROCESSING = Dict[str, Union[str, List[int], int, float, Dict[str, int]]]

    def __init__(self, special_processing_dict:SPECIAL_PROCESSING, square_type:str):
        """Constructor
            Arguments:
                special_processing_dict - the JSON specialProcessing section of a border or occupation square as a dictionary
                square_type - "occupation" or "border"
        """
        self._special_processing_dict = special_processing_dict
        #
        # mandatory elements
        #
        self._square_type = square_type     # "occupation" or "border"
        self._processing_type = special_processing_dict["type"]     # must be one of the above types
        #
        # optional elements which are type-dependent
        # an 'amount' can be a numeric type (int or float), a string or a dictionary
        #
        self._amount = 0            # a discrete money amount which may be salary or cash depending on type
        self._amount_dict = None
        self._amount_str = ""
        amt = special_processing_dict.get('amount', 0)    # 'amount' is optional
        if isinstance(amt, int) or isinstance(amt, float):
            self._amount = amt
        elif isinstance(amt, str):
            self._amount_str = amt
        else:
            self._amount_dict = amt    # how a money amount is calculated
        
        self._dice = special_processing_dict.get('dice', 0)                 # number of dice used to calculate some amount
        self._penalty = special_processing_dict.get('penalty', 0)           # a loss quantity (hearts or stars)
        self._limit = special_processing_dict.get('limit', None)            # a limiting factor, usually "salary" (in 1000s)
        
        # square number to advance to, relative to the overall game layout or occupation
        self._next_square = special_processing_dict.get('next_square', None)    

        self._destination = special_processing_dict.get('destinationOccupation', None)  # the name of a destination occupation
        self._percent = special_processing_dict.get('percent', 0.0)                     # a percent amount of salary or cash depending on type
        self._must_roll = special_processing_dict.get('must_roll', [])                  # a list of numbers that must be rolled in order to leave this square
        self._require_doubles = special_processing_dict.get('require_doubles', 0)==1
        self._hearts = special_processing_dict.get('hearts', [])                        # used for holiday type, a 2-element list
    
    @property
    def square_type(self):
        return self._square_type
    
    @property
    def processing_type(self):
        return self._processing_type
    
    @property
    def amount(self):
        return self._amount
    
    @property
    def amount_dict(self):
        return self._amount_dict
    
    @property
    def amount_str(self):
        return self._amount_str
    
    @property
    def destination(self):
        return self._destination
    
    @property
    def dice(self):
        return self._dice
    
    @property
    def limit(self):
        return self._limit
    
    @property
    def next_square(self):
        return self._next_square
    
    @property
    def penalty(self):
        return self._penalty
    
    @property
    def percent(self):
        return self._percent
    
    @property
    def must_roll(self):
        return self._must_roll
    
    @property
    def require_doubles(self):
        return self._require_doubles
    
    @property
    def hearts(self) -> list:
        return self._hearts
    
    def __str__(self):
        return str(self._special_processing_dict)
    
    def to_JSON(self):
        return json.dump(self._special_processing_dict, indent=2)



'''
Created on Aug 21, 2022

@author: don_bacon
'''
from game.careersObject import CareersObject
from game.boardLocation import BoardLocation

class CommandResult(CareersObject):
    """Encapsulates the result of a command executed by a game square or the CareersGameEngine
    
    """
    SUCCESS = 0         # Successful result
    ERROR = 1           # There was an error. message contains the error message
    TERMINATE = 2       # terminate the game
    EXECUTE_NEXT = 3    # successful, and execute the next_action for the current player
    NEED_PLAYER_CHOICE = 4  # successful, but need player choice as to what to do next

    def __init__(self, return_code:int, message:str, done_flag:bool, next_action=None, board_location:BoardLocation=None, exception:Exception=None):
        """Constructor, baby.
            Arguments:
                return_code - integer return code:
                    SUCCESS = 0  Successful result
                    ERROR = 1  There was an error. message contains the error message
                    TERMINATE = 2  terminate the game
                    EXECUTE_NEXT = 3  success, and execute the next_action for the current player
                message - a message string to be displayed to the player
                done_flag - if  True, this player's turn is completed, False otherwise.
                next_action - next action to perform for this player, default is None. 
                        This will be in the format of an executable command, for example "roll" as this is
                        passed to the game engine execute_command() method.
                board_location - the board location (of the current player) after the command is executed.
                exception - the Exception instance if the command raised an exception, default is None
        """
        self._return_code = return_code
        self._message = message
        self._done_flag = done_flag
        self._exception = exception
        self._next_action = next_action
        self._board_location = board_location
    
    @property
    def return_code(self):
        return self._return_code
    @return_code.setter
    def return_code(self, value):
        self._return_code = value
    
    @property
    def message(self):
        return self._message
    @message.setter
    def message(self, amessage):
        self._message = amessage
    
    @property
    def done_flag(self):
        return self._done_flag
    @done_flag.setter
    def done_flag(self, value):
        self._done_flag = value
    
    @property
    def exception(self):
        return self._exception
    @exception.setter
    def exception(self, value):
        self._exception = value
        
    @property
    def next_action(self):
        return self._next_action
    @next_action.setter
    def next_action(self, value):
        self._next_action = value
        
    @property
    def board_location(self):
        return self._board_location
    @board_location.setter
    def board_location(self, value):
        self._board_location = value
        
    def is_successful(self):
        return True if self.return_code == CommandResult.SUCCESS else False
        
    def to_JSON(self):
        jstr = f'{{\n  "return_code" : "{self.return_code}",\n  "done_flag" : "{self.done_flag}",\n  "message" : "{self.message}",\n'
        if self.next_action is not None:
            jstr += f'  "next_action" : "{self.next_action}",\n'
        if self.exception is not None:
            jstr += f'  "exception : "{str(self.exception)}",\n'
        if self.board_location is not None:
            bl = self.board_location.to_JSON()
            jstr += f'"board_location" : {bl}\n'
        jstr += "}"
        return jstr
        


'''
Created on Aug 9, 2022

@author: don_bacon
'''

from enum import Enum
class ExperienceCardType(Enum):
    """
    Experience card types
    """
    FIXED = 1
    SINGLE_DIE_WILD = 2
    DOUBLE_DIE_WILD = 3
    TRIPLE_DIE_WILD = 4
        
    def describe(self):
        # self is the member here
        return self.name, self.value
        


'''
Created on Aug 12, 2022

@author: don_bacon
'''

from game.gameSquare import GameSquare
from game.player import Player
from game.commandResult import CommandResult
import json

class BorderSquare(GameSquare):
    """Encapsulates a Careers game border (non-occupation) square.
    Border squares are numbered consecutively starting with 0 (Payday by convention).
    They also have a name which may not be unique. For example, there are 12 squares named "Opportunity"
    The border square "type" is enumerated in the gameLayout JSON as "types_list
    """

    types_list = ["corner_square", "opportunity_square", "danger_square", "travel_square", "occupation_entrance_square", "action_square"]


    def __init__(self, border_square_dict, game=None):
        """Create a BorderSquare instance.
            Arguments:
                border_square_dict - the dictionary defining this BorderSquare. This would be an element of the game layout.
                game - a CareersGame instance

        """
        super().__init__("Border", name=border_square_dict['name'], number= border_square_dict['number'], \
                         text=border_square_dict['text'], special_processing_dict=border_square_dict['specialProcessing'], game=game)
        
        self._game_square_dict = border_square_dict
        self._square_type = border_square_dict['type']
        self.action_text = border_square_dict.get('action_text', None)
        
    @property
    def square_type(self):
        return self._square_type
    
    @square_type.setter
    def square_type(self, value):
        self._square_type = value
        
    @property
    def game_square_dict(self):
        return self._game_square_dict
    
    def execute(self, player:Player) -> CommandResult:
        """Execute the actions associated with landing on this BorderSquare. Overrides GameSquare.execute().
            Assumes the Player's current location is this BorderSquare.
            Arguments:
                player - current Player
            Returns: CommandResult
        """
        careersGame = self._careersGame
        if self.name == 'Opportunity':
            deck = careersGame.opportunities
            card = deck.draw()
            player.my_opportunity_cards.append(card)
            return CommandResult(CommandResult.SUCCESS, f'Added Opportunity: {str(card)}', True)
        
        elif self.square_type == 'travel_square':
            #
            # advance to the next travel_square and roll again
            # 
            #
            next_square_number = self._careersGame.find_next_border_square(self.number, 'travel_square')
            game_square = careersGame.game_board.get_square(next_square_number)
            next_action = f'goto {next_square_number};roll'    # player.board_location set by 'goto' command

            result = CommandResult(CommandResult.SUCCESS, f'Advance to square {next_square_number}, {game_square.name} and roll again', False)
            result.next_action = next_action
            #result.board_location = player.board_location
            return result
        
        elif self.square_type == 'occupation_entrance_square':
            #
            # can enter if landed here using an Opportunity, else turn is over
            #
            if player.opportunity_card is not None:
                if player.opportunity_card.opportunity_type  == 'occupation' and player.opportunity_card.name == self.name:
                    next_action = f'enter {self.name}'
                    result = CommandResult(CommandResult.SUCCESS, f'Entering {self.name}', False, next_action=next_action)
            else:
                result = CommandResult(CommandResult.SUCCESS, "", True)
            return result
        elif self.square_type == 'action_square':
            if self.name == 'BuyInsurance':         # nothing to do, buy_insurance is a separate command
                return CommandResult(CommandResult.NEED_PLAYER_CHOICE, "", False)
            else:
                #
                # return the JSON dumps of this square as landing here requires a choice by the player
                # to buy hearts, stars, experience cards or to gamble
                # The UI will send back the appropriate command if the player wants to execute the square
                # this will include the choice amount.
                #
                message = self.to_JSON()
                return CommandResult(CommandResult.NEED_PLAYER_CHOICE, message, False)
        elif self.square_type == 'corner_square':
                # check special processing type because corner square names are edition-dependent, specialProcessing type is independent of the edition
    
                sp_type = self.special_processing.processing_type
                if sp_type == "unemployment":
                    player.is_unemployed = True
                    self.update_board_location(player)
                    result = CommandResult(CommandResult.SUCCESS, f'Player {player.player_initials}: {self.action_text}', True)
                elif sp_type == "hospital":
                    player.is_sick = True
                    self.update_board_location(player)
                    result = CommandResult(CommandResult.SUCCESS, f'Player {player.player_initials}: {self.action_text}', True)                  
                elif sp_type == "payday":
                    salary = player.salary
                    # if I am on Payday I get double salary
                    # don't update the player's board_location 
                    how = "passed" 
                    if player.board_location.border_square_number == 0: 
                        salary += salary
                        how = "landed on"
                    player.cash += salary
                    player.laps += 1
                    result = CommandResult(CommandResult.SUCCESS, f'Player {player.player_initials} {how} {self.name}\n{self.action_text}', True)
                elif sp_type == "holiday":    # a.k.a. Spring Break, Holiday
                    # the number of hearts you get is configured in the specialProcessing section 
                    nhearts = self.special_processing.hearts[0] if player.on_holiday else self.special_processing.hearts[1]
                    player.on_holiday = True
                    player.add_hearts(nhearts)
                    self.update_board_location(player)
                    result = CommandResult(CommandResult.SUCCESS, f'Player {player.player_initials} {self.action_text}\n collect {nhearts} hearts', True)
                    
                return result
        else:
            pass
        
        result = CommandResult(CommandResult.SUCCESS, f'{self.square_type} {self.name} execute not yet implemented', False)   #  TODO
        return result
    
    def execute_special_processing(self, player:Player):
        """Invokes the specialProcessing section of this border square for a player.
            This method is called as exit processing for a square, whereas execute() is called when a player lands on a border square.
            
            TODO - finish this, for now return SUCCESS
        """
        result = CommandResult(CommandResult.SUCCESS, f'{self.square_type} {self.name} execute_special_processing not yet implemented', False)
        return result
    
    def update_board_location(self, player:Player):
        player.board_location.border_square_number = self.number
        player.board_location.border_square_name = self.name
        player.board_location.occupation_name = None     
        
    def to_JSON(self):
        return json.dumps(self.game_square_dict)


    


'''
Created on Aug 5, 2022

@author: don_bacon
'''
from game.successFormula import SuccessFormula
from game.careersObject import CareersObject
from game.boardLocation import BoardLocation
from game.opportunityCard import OpportunityCard
from game.experienceCard import ExperienceCard

from datetime import datetime
import json

class Player(CareersObject):
    
    def __init__(self, number=0, name="Player", salary=2000, cash=2000, initials="XXX"):
        self._player_name = name
        self._player_initials = initials            # unique initials - no player can have the same initials
        self._salary_history = [salary]             # list of salaries the player has attained
        self._salary = salary                       # my current salary
        self._cash = cash                           # cash on hand
        self._success_formula = None         # my SuccessFormula
        self._happiness = [0]                # record of happiness (hearts) earned. Cumulative amounts, total is happiness[-1]
        self._fame = [0]                     # record of fame (stars) earned. Cumulative amounts, total is fame[-1]
        
        self._number = number               # my player number, values 0 to #players-1
        self._my_experience_cards = []      # list of ExperienceCards this player holds
        self._my_opportunity_cards = []     # list of OpportunityCards this player holds
        
        # dict where key is the degree program and the values the number of degrees attained
        # when number of degrees in any program >= 4, the can_retire flag is automatically set
        self._my_degrees = dict()
        # always start on Payday corner square
        self._board_location = BoardLocation(border_square_number=0, border_square_name="Payday", occupation_name=None, occupation_square_number=0 )
        self._can_retire = False
        self._is_insured = False
        self._is_unemployed = False         # True when player lands on (or is sent to) Unemployment
        self._is_sick = False               # True when player lands on Hospital
        self._on_holiday = False            # True when the player lands on Holiday/Spring Break
        self._lose_turn = False             # If True the player loses their next turn. This is automatically reset when the turn is skipped.
        
        # dict where key is occupation name, value is the number of completed trips
        # when completed trips >= 3, the can_retire flag is automatically set
        self._occupation_record = dict()
        
        # player loan obligations are indexed by player_number: loans[player_number] = <loan amount>
        self._loans = {}
        
        self._opportunity_card = None   # the OpportunityCard instance of the card currently in play, or None otherwise
        self._experience_card = None    # the ExperienceCard instance of the card currently in play, or None otherwise
        self._laps = 0                  # the number of times player has passed or landed on Payday
        self._can_bump = []             # the players I can currently Bump
        
    @property
    def player_name(self):
        """Get the player's name."""
        return self._player_name
    
    @player_name.setter
    def player_name(self, value):
        self._player_name = value
        
    @property
    def player_initials(self):
        return self._player_initials
    
    @property
    def salary(self):
        return self._salary
    
    @salary.setter
    def salary(self, newSalary):
        self._salary = newSalary
        self._salary_history.append(newSalary)
        
    @property
    def salary_history(self):
        return self._salary_history
    
    @property
    def cash(self):
        return self._cash
    
    @cash.setter
    def cash(self, value):
        self._cash = value
    
    @property
    def success_formula(self):
        return self._success_formula
    
    @success_formula.setter
    def success_formula(self, value):
        self._success_formula = value
        
    @property
    def number(self):
        return self._number
    
    @number.setter
    def number(self, value):
        self._number = value
        
    @property
    def my_degrees(self):
        return self._my_degrees
    
    @property
    def occupation_record(self):
        return self._occupation_record
    
    @property
    def can_retire(self):
        return self._can_retire
    
    @can_retire.setter
    def can_retire(self, value:bool):
        self._can_retire = value
    
    @property
    def is_insured(self):
        return self._is_insured
    
    @is_insured.setter
    def is_insured(self, value:bool):
        self._is_insured = value
    
    @property
    def happiness(self):
        return self._happiness[-1]
    
    @property   
    def fame(self):
        return self._fame[-1]
    
    @property
    def board_location(self):
        return self._board_location
    
    @board_location.setter
    def board_location(self, value):
        self._board_location = value


    def current_border_square_number(self):
        return self.board_location.border_square_number
        
    def current_border_square_name(self):
        return self.board_location.border_square_name

    def current_occupation_name(self):
        return self.board_location.occupation_name
    
    def current_occupation_square_number(self):
        return self.board_location.occupation_square_number
    
    @property
    def my_opportunity_cards(self):
        return self._my_opportunity_cards
    
    @property
    def opportunity_card(self) -> OpportunityCard:
        """Returns the OpportunityCard currently in play, or None if no card is in play
        """
        return self._opportunity_card
    
    @opportunity_card.setter
    def opportunity_card(self, value:OpportunityCard):
        """Sets the OpportunityCard currently in play. After use, set to None.
        """
        self._opportunity_card = value
        
    def add_opportunity_card(self, thecard:OpportunityCard):
        """Add a single OpportunityCard to my deck
        """
        self._my_opportunity_cards.append(thecard)
    
    @property
    def my_experience_cards(self):
        return self._my_experience_cards
        
    @property
    def experience_card(self) -> ExperienceCard:
        """Returns the ExperienceCard currently in play, or None if no card is in play
        """
        return self._experience_card
    
    @experience_card.setter
    def experience_card(self, value:ExperienceCard):
        """Sets the ExperienceCard currently in play
        """
        self._experience_card = value
        
    def add_experience_card(self, thecard:ExperienceCard):
        """Add a single ExperienceCard to my deck
        """
        self._my_experience_cards.append(thecard)
    
    @property
    def is_unemployed(self):
        return self._is_unemployed
    
    @is_unemployed.setter
    def is_unemployed(self, value):
        self._is_unemployed = value
    
    @property
    def is_sick(self):
        return self._is_sick
    
    @is_sick.setter
    def is_sick(self, value):
        self._is_sick = value
        
    @property
    def on_holiday(self):
        return self._on_holiday
    
    @on_holiday.setter
    def on_holiday(self, value):
        self._on_holiday = value
        
    @property
    def lose_turn(self):
        return self._lose_turn
    
    @lose_turn.setter
    def lose_turn(self, value):
        self._lose_turn = value
        
    @property
    def laps(self):
        return self._laps
    
    @laps.setter
    def laps(self, value):
        self._laps = value
    
    @property
    def loans(self):
        return self._loans
    
    @property
    def can_bump(self):
        return self._can_bump       #a list of Player
    
    @can_bump.setter
    def can_bump(self, other_players):
        self._can_bump = other_players    #a list of Player
    
    def get_total_loans(self):
        total = 0
        for v in self._loans.values():
            total += v
        return total
    
    def add_loan(self, amt, player_number):
        if player_number in self._loans:
            self._loans[player_number] = self._loans[player_number] + amt
        else:
            self._loans[player_number] = amt
    
    def get_opportunity_cards(self) -> list:
        """Returns a of dict of Opportunity cards indexed by number
            This method used to display the cards to a player for selection.
            Return format is:
            {  4 :  { "quantity" : 3,  "text" : "card text", "card" : <card instance>} }
        """
        cards_dict = {}
        for card in self.my_opportunity_cards:
            key = card.number
            if key in cards_dict:
                cd = cards_dict[key]
                cd["quantity"] = cd["quantity"] + 1
            else:
                cards_dict[key] = {"quantity" : 1, "text" : card.text, "card" : card}
        return cards_dict
    
    def get_experience_cards(self):
        """Returns a of dict of Experience cards indexed by number.
            This method used to display the cards to a player for selection.
            Return format is:
            {  4 :  { "quantity" : 3,  "spaces" : "4", "card" : <card instance>} } if spaces > 0,  or "spaces" : <card type>
        """        
        cards_dict = {}
        for card in self.my_experience_cards:
            key = card.number
            if key in cards_dict:
                cd = cards_dict[key]
                cd["quantity"] = cd["quantity"] + 1
            else:
                spaces = card.spaces
                if spaces > 0:
                    cards_dict[key] = {"quantity" : 1, "spaces" : str(spaces), "card" : card}
                else:
                    cards_dict[key] = {"quantity" : 1, "spaces" : card.type, "card" : card}
        return cards_dict
    
    def used_opportunity(self):
        """The player has used the saved Opportunity card. Set it to None and remove from their deck
        """
        if self.opportunity_card is not None:
            self.my_opportunity_cards.remove(self._opportunity_card)
            self._opportunity_card = None
    
    def used_experience(self):
        """The player has used the saved Experience card. Set it to None and remove from their deck
        """
        if self.experience_card is not None:
            self.my_experience_cards.remove(self.experience_card)
            self._experience_card = None
                
    def add_degree(self, degree_program):
        if degree_program in self.my_degrees:
            count = self.my_degrees[degree_program]
            # max of 4 degrees in any degree program
            count += 1
            if count <= 4:
                self._my_degrees[degree_program] = count
            if count >= 4:
                self.can_retire = True
        else:
            self._my_degrees[degree_program] = 1
    
    def add_occupation(self, occupation_name):
        if occupation_name in self.occupation_record:
            count = self._occupation_record[occupation_name] + 1
            self._occupation_record[occupation_name] = count
            if count >= 3:
                self.can_retire = True
        else:
            self._occupation_record[occupation_name] = 1
            
        
    def add_hearts(self, nhearts):
        self._happiness.append(self.happiness + nhearts)
    
    def add_stars(self, nstars):
        self._fame.append(self.fame + nstars)
        
    def add_cash(self, money):  # could be a negative amount
        self.cash = self.cash + money
    
    def add_to_salary(self, money): #  again, could be negative amount
        self.salary = self.salary + money
        self._salary_history.append(self.salary)
        
    def total_points(self):
        return self.happiness  + self.fame + self.cash_points() - self.loan_points() 
    
    def cash_points(self):
        """Cash points is the amount of cash/1000
        """
        return self.cash // 1000
    
    def loan_points(self):
        return self.get_total_loans() // 1000
    
    def is_complete(self):
        """Returns True if this players's total points are >= game total points, False otherwise
    
        """
        return  self.total_points() >= self.success_formula.total_points()

    def save(self, gameId=None):
        """Persist this player's state to a JSON file.
        File name is "player_" + player_initials + yyyy-mm-dd_hhmmss + _state.json"
        
        TODO finish this
        """
        today = datetime.now()
        fdate = '{0:d}-{1:02d}-{2:02d}_{3:02d}{4:02d}{5:02d}'.format(today.year,today.month, today.day, today.hour, today.minute, today.second)
        if gameId is None:
            filename = f'player_{self._player_initials}_{fdate}_state.json '
        else:
            filename = f'{gameId}_player_{self._player_initials}_{fdate}_state.json'
        #
        # save in jsonpickle format
        #
        jtxt = self.json_pickle()
        
        return filename
    
    def player_info(self, include_successFormula=False):
        v = self.get_total_loans()
        fstring = f'''salary:{self.salary}, Cash: {self.cash},  Fame: {self.fame}, Happiness: {self.happiness}, 
Insured: {self.is_insured}, Unemployed: {self.is_unemployed}, Sick: {self.is_sick}'''
        
        if include_successFormula:
            fstring = f'{fstring}\nSuccess Formula: {self.success_formula}'
        if v > 0:
            fstring = f'{fstring}\nloans: {v}'
        return fstring
    
    def get_current_location(self) -> BoardLocation:
        """Gets the location of this player on the board.
            Returns: current_board_location
        
        """
        return  self.current_board_location
    
    def info(self):
        return  f' "name" : "{self.player_name}",  "number" : "{self.number}",  "initials" : "{self.player_initials}"'
    
    def __str__(self):
        fstring = f'Cash: {self.cash}  Fame: {self.fame}  Happiness: {self.happiness}'
        return f'{self.number}. {self.player_name} ({self.player_initials}) : salary:{self.salary}\n{fstring}\nSuccess Formula: {self.success_formula}'
    
    def to_dict(self):
        pdict = {"name" : self.player_name, "number" : self.number, "initials" : self.player_initials}
        pdict['successFormula'] = self._success_formula.to_dict()
        points = self.total_points()
        pdict['score'] = {"cash":self.cash, "fame":self.fame, "happiness":self.happiness, "total_points":points, "is_insured":self.is_insured}
        pdict['loans'] = self.loans
        pdict['board_location'] = self.board_location.to_dict()
        pdict['is_sick'] = self.is_sick
        pdict['is_unemployed'] = self.is_unemployed
        pdict['occupation_record'] = self.occupation_record
        
        return pdict

    def to_JSON(self):
        return json.dumps(self.to_dict(), indent=2)

    
if __name__ == '__main__':
    player = Player(0, name='Don', initials='DWB')
    sf = SuccessFormula(stars=40, hearts=10, cash=50)
    player.success_formula = sf
    
    print(str(player))
    print(repr(player))
    
    player.add_cash(5000)
    player.add_hearts(3)
    player.add_stars(6)
    print("\n" + player.player_info(include_successFormula=False) )
    player.add_to_salary(2000)
    print("\n" + player.player_info(include_successFormula=True)  )
    
    print("\njson_pickle:")
    print(player.json_pickle())
    print("\nto_JSON:")
    print(player.to_JSON())
    


'''
Created on Aug 13, 2022

@author: don_bacon
'''

import jsonpickle
class CareersObject(object):
    """Base class for object that have a JSON representation.
    """

    def __init__(self, params):
        '''
        Constructor
        '''
        pass
    
    def to_JSON(self):
        """Returns my JSON representation. Abstract class.
        """
        return None
    
    def json_pickle(self):
        """A complete JSON representation of this object using jsonpickle.
        """
        return jsonpickle.encode(self, indent=2)
        
    def __repr__(self):
        return self.json_pickle()
    
    def __str__(self):
        return self.to_JSON()
    


'''
Created on Sep 11, 2022

@author: don_bacon
'''

from game.careersGame import CareersGame
from game.commandResult import CommandResult
from game.player import Player
from game.gameUtils import GameUtils
import joblib

class GameEngineCommands(object):
    """Implementations of CareersGameEngine commands.
    
    """
    COMMANDS = ['add', 'bankrupt', 'bump', 'buy', 'create', 'done', 'end', 'enter', 
                'game_status', 'goto', 'list', 'load', 'next', 'pay', 'quit', 'retire', 
                'roll', 'save', 'saved', 'start', 'status', 'transfer', 'use', 'use_insurance', 
                'where', 'who']
    

    def __init__(self, thegame:CareersGame, logfile_pointer):
        self._careersGame = thegame
        self._game_state = self._careersGame.game_state
        self.fp = logfile_pointer
        self._trace = True          # traces the action by describing each step and logs to a file
    
    @property 
    def careersGame(self) -> CareersGame:
        return self._careersGame
    
    @careersGame.setter
    def careersGame(self, thegame:CareersGame):
        self._careersGame = thegame
        
    @property
    def trace(self):
        return self._trace
    
    @trace.setter
    def trace(self, value):
        self._trace = value
        
    def can_player_move(self, player:Player) -> tuple[bool,CommandResult] :
        """Determine if a Player can move from the Hospital or Unemployment
            Returns: a 2-element tupple consisting of a boolean (the player can move or not)
            and a CommandResult.
        """
        if player.is_unemployed or player.is_sick:
            bs_name = player.current_border_square_name()
            # bs_number = player.current_border_square_number
            game_square = self.careersGame.find_border_square(bs_name)
            #
            # let the square determine if the player can move or not
            # as that's encoded in the specialProcessing
            #
            result = game_square.execute_special_processing(player)
        else:
            result = CommandResult(CommandResult.SUCCESS, "", True)
            
        can_move = result.is_successful()
        return can_move, result

    @staticmethod
    def parse_command_string(txt, addl_args=[]) -> CommandResult:
        """Parses a command string into a string that can be evaluated with eval()
            Returns: if return_code == 0, a CommandResult with commandResult.message as the string to eval()
                else if return_code == 1, commandResult.message has the error message
        """
        command_args = txt.split()
            
        command = command_args[0]
        if not command in GameEngineCommands.COMMANDS:
            return CommandResult(CommandResult.ERROR,  f'Invalid command: {command}',  False)
        if len(command_args) > 1:
            args = command_args[1:]
            command = command + "("
            for arg in args:
                if arg.isdigit():
                    command = command + arg + ","
                else:
                    command = command + f'"{arg}",'
        
            command = command[:-1]    # remove the trailing comma
        else:
            command = command + "("
            
        if addl_args is not None and len(addl_args) > 0:
            for arg in addl_args:
                command = command + f'"{arg}",'
            command = command[:-1]
        command += ")"
        
        return CommandResult(CommandResult.SUCCESS, command, False)
    
    @staticmethod
    def list(player, what) ->CommandResult:
        """List the Experience or Opportunity cards held by the current player
            Arguments: what - 'experience', 'opportunity', or 'all'
            Returns: CommandResult.message is the stringified list of str(card).
                For Opportunity cards this is the text property.
                For Experience cards this is the number of spaces (if type is fixed), otherwise the type.
            
        """
        message = ""
        listall = (what.lower() == 'all')

        if what.lower().startswith('opportun') or listall:
            if len(player.my_opportunity_cards) == 0:
                message += "No Opportunity cards\n"
            else:
                n = 1
                for card in player.my_opportunity_cards:
                    num = card.number
                    message += f'{n}.  {num}: {str(card)}\n'
                    n += 1
                    
        if what.lower().startswith('exp') or listall:
            if len(player.my_experience_cards) == 0:
                message += "\nNo Experience cards"
            else:
                n = 1
                for card in player.my_experience_cards:
                    message += f'{n}.  {num}: {str(card)}\n'
                    n+= 1
            
        result = CommandResult(CommandResult.SUCCESS, message, False)
        return result   
    
    def save_game(self, gamefile_base_name:str, game_id:str, how='json') -> CommandResult:
        """Save the complete serialized game state so it can be restarted at a later time.
            Arguments:
                how - serialization format to use: 'json', 'jsonpickle' or 'pkl' (pickle)
            NOTE that the game state is automatically saved in pkl format after each player's turn.
            NOTE saving in JSON format saves only the GameState; pkl and jsonpickle persist CareersGame
        """
        extension = 'pkl' if how=='pkl' else 'json'
        filename = f'{gamefile_base_name}.{extension}'      # folder/filename
        if how == 'json':
            jstr = f'{{\n  "game_id" : "{game_id}",\n'
            jstr += f'  "gameState" : '
            jstr += self.game_state.to_JSON()
            jstr += "}\n"

            with open(filename, "w") as fp:
                fp.write(jstr)
            fp.close()
        elif how == 'jsonpickle':
            jstr = self._careersGame.json_pickle()
            with open(filename, "w") as fp:
                fp.write(jstr)
            fp.close()
        else:
            result = joblib.dump(self._careersGame, filename)   # returns a list, as in  ['/data/games/ZenAlien2013_20220909-124721-555368-33134.pkl']
            filename = f'{result}'
        
        self.log(f'game saved to {filename}')
        return CommandResult(CommandResult.SUCCESS, filename, True)


    def log(self, message):
        """Write message to the log file.
            TODO - refactor to use python loging
        """
        msg = GameUtils.get_datetime() + f'  {message}\n'
        if self.fp is not None:     # may be logging isn't initialized yet or logging option is False
            self.fp.write(msg)
        if self.trace:
            print(msg)
    



from threading import Lock
import sys

class Environment(object):
    _environ = None
    _lock = Lock()
    
    def __init__(self, package_name=None):
        """Initialize the running Environment by setting global environment variables
        
        """
        self.name = package_name
        self.resource_base = '/Compile/careers/resources'
        self.package_name = package_name
        self.resources = {}
        self.resources['game'] = self.resource_base + '/game'
        self.resources['server'] = self.resource_base + '/server'
        self.items = {}
        
    def __repr__(self):
        return '<Environment>'
    
    
    def get_resource_folder(self, package=None):
        if package is None:
            return self.resource_base
        if package in self.resources:
            return self.resources[package]
        else:
            print(f'Invalid package {package}', file=sys.stderr)
            return None

    def add_item(self, key, value):
        self.items[key] = value
    
    def get_items(self):
        return self.items
    
    def get_item(self, key):
        return self.items[key]

    @staticmethod
    def get_environment():
        env = None
        Environment._lock.acquire()
        if Environment._environ is not None:
            env = Environment._environ
        else:
            Environment._environ = Environment('dwbzen')
            env = Environment._environ
        Environment._lock.release()
        return env




# ------------------------------------------------------------------------------
# Name:          careersGameEngine.py
# Purpose:       CareersGameEngine class.
#
#                CareersGameEngine executes commands against player-board position
#
# Authors:      Donald Bacon
#
# Copyright:    Copyright (c) 2022 Donald Bacon
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------

from game.careersGame import CareersGame
from game.commandResult import CommandResult
from game.player import  Player
from game.opportunityCard import OpportunityCard
from game.experienceCard import ExperienceCard
from game.boardLocation import BoardLocation
from game.successFormula import SuccessFormula
from game.borderSquare import BorderSquare
from game.occupationSquare import OccupationSquare
from game.gameSquare import GameSquare
from game.gameEngineCommands import GameEngineCommands
from game.gameUtils import GameUtils

from datetime import datetime
import random
from typing import Optional, List


class CareersGameEngine(object):
    """CareersGameEngine executes the action(s) associated with each player's turn.
    Valid commands + arguments:
        command :: <roll> | <use> | <retire> | <bump> | <bankrupt> | <list> | <status> | <quit> | <done> | <end game> |
                   <saved games> | <save> | <load> | <query> | <enter> | <goto> | <add> | <use insurance> | <add degree> |
                   <pay> | <transfer> | <game_status> | <create> | <start> | <buy>
        <use> :: "use"  <card-type>
        <card_type> :: "opportunity" | "experience" 
        <roll> :: "roll"                        ;roll 1 or 2 dice depending on where the player is on the board
        <retire> :: "retire"                    ;immediate go to retirement square (Spring Break, Holiday)
        <bump> :: "bump" player_initials        ;bump another player, who must be on the same square as the bumper
        <bankrupt> :: "bankrupt"                ;declare bankruptcy
        <list> :: "list"  <card_type> | "occupations"           ;list my opportunities or experience cards, or occupations completed
        <status> :: "status"                    ;display my cash, #hearts, #stars, salary, total points, and success formula
        <quit> :: "quit" player_initials        ;current player leaves the game, must include initials
        <done> :: "done" | "next"               ;done with my turn - next player's turn
        <end game> :: "end game"                ;saves the current game state then ends the game
        <saved games> :: "saved games"          ;list any saved games by date/time and gameID
        <save> :: "save" <how>                  ;saves the current game state to a file in the specified format
        <how> ::  "json" | "pkl"
        <load> :: "load" game-id                ;load a game and start play with the next player
        <query> :: "where" <who>                ;gets info on a player's current location on the board
        <who> :: "am I" | "is" <playerID>
        <playerID> :: player_name | player_initials
        <enter> :: "enter" <occupation_name> [<square_number>]                 ;enter occupation at occupation square square_number
        <goto> :: "goto" <square_number>                                       ;go to border square square_number
        <add> :: "add player" player_name player_initials cash stars hearts    ;adds a new player to the game
        <use insurance> :: "use_insurance"
        <add degree> :: "add degree" <degree program>
        <degree program> :: See collegeDegrees-<edition name>.json "degreePrograms"
        <pay> :: "pay" amount                ;current player makes a payment associated with their current board location
        <transfer> :: "transfer" quantity ("cash" | "opportunity" | "experience) player_number
        <game_status> :: "game_status"
        <create> :: "create" <edition> <game_type> points  [id]      ;create a new CareersGame with specified total points or total time (in minutes) and optional gameId
        <edition> :: "Hi-Tech"    ;supports multiple editions
        <game_type> ::  'points' | 'timed'
        <start> :: "start"                    ;starts a newly created CareersGame
        <buy>  :: "buy"  ( "hearts" | "stars" | "experience" | "opportunity" ) quantity cash_amount  ;buy some number of items for the cash_amount provided
    """
    
    
    def __init__(self):
        '''
        Constructor
        '''
        self._fp = None             # logging file pointer
        self._careersGame = None    # create a new CareersGame with create()
        self._trace = True          # traces the action by describing each step and logs to a file
        self._start_date_time = datetime.now()
        self._gameId = None
        self._master_id = None      # provided by the UI
        self._current_player = None
        self._admin_player = Player(number=-1, name='Administrator', initials='admin')
        self._gameEngineCommands = None     # no CareersGame yet
    
    @property
    def fp(self):
        return self._fp
    
    @fp.setter
    def fp(self, value):
        self._fp = value
    
    @property
    def logfile_name(self):
        return self._logfile_name
    
    @property
    def logfile_path(self):
        return self._logfile_path
    
    @property
    def gameId(self):
        return self._gameId
    
    @property
    def master_id(self):
        return self._master_id
    
    @property
    def trace(self):
        return self._trace
    
    @trace.setter
    def trace(self, value):
        self._trace = value
        
    @property
    def game_state(self):
        return self._game_state
    
    @property
    def careersGame(self):
        return self._careersGame
    
    def log(self, message):
        """Write message to the log file.
            TODO - refactor to use python loging
        """
        msg = GameUtils.get_datetime() + f'  {message}\n'
        if self.fp is not None:     # may be logging isn't initialized yet or logging option is False
            self.fp.write(msg)
        if self.trace:
            print(msg)

    def execute_command(self, command:str, aplayer:Player, args:list=[]) -> CommandResult:
        """Executes command(s) for a given Player
            Arguments:
                command - the command name, for example "roll". Multiple commands are separated by a semicolon.
                args - a possibly empty list of additional string arguments
                player - a Player reference. If none, admin_player is used.
            Returns: a CommandResult object. The player's current board_location is always returned in the CommandResult
            See game.CommandResult for details
        
        """
        player = aplayer
        if aplayer is None:
            player = self._admin_player
        
        commands = command.split(';')
        messages = ""
        for command in commands:
            self.log(f'{player.player_initials}: {command}')
            if command is None or len(command) == 0:
                return CommandResult(CommandResult.SUCCESS, "", False)
            cmd_result = self._evaluate(command, args)
            
            board_location = player.board_location    # current board location AFTER the command is executed
            self.log(f'  {player.player_initials} results: {cmd_result.return_code} {cmd_result.message}\n{board_location}')
            cmd_result.board_location = board_location
            messages += cmd_result.message + "\n"
        
        cmd_result.message = messages
        return cmd_result
        
    def _evaluate(self, commandTxt, args=[]) -> CommandResult:
        """Evaulates a command string with eval()
            Arguments:
                commandTxt - the command name + any arguments to evaluate.
                args - an optional list of additional arguments
            Returns - a CommandResult
        """
        command_result = GameEngineCommands.parse_command_string(commandTxt, args)
        if command_result.return_code != CommandResult.SUCCESS:     # must be an error
            return command_result
        
        command = "self." + command_result.message
        command_result = None
        print("execute " + command)
        try:
            command_result = eval(command)
        except Exception as ex:
            command_result = CommandResult(CommandResult.ERROR,  f'"{command}" : Invalid command format or syntax\n{str(ex)}',  False, exception=ex)
            #raise ex
        return command_result
        
    def _parse_command_string(self, txt, addl_args=[]) -> CommandResult:
        """Parses a command string into a string that can be evaluated with eval()
            Returns: if return_code == 0, a CommandResult with commandResult.message as the string to eval()
                else if return_code == 1, commandResult.message has the error message
        """
        command_args = txt.split()
            
        command = command_args[0]
        if not command in CareersGameEngine.COMMANDS:
            return CommandResult(1,  f'Invalid command: {command}',  False)
        if len(command_args) > 1:
            args = command_args[1:]
            command = command + "("
            for arg in args:
                if arg.isdigit():
                    command = command + arg + ","
                else:
                    command = command + f'"{arg}",'
        
            command = command[:-1]    # remove the trailing comma
        else:
            command = command + "("
            
        if addl_args is not None and len(addl_args) > 0:
            for arg in addl_args:
                command = command + f'"{arg}",'
            command = command[:-1]
        command += ")"
        
        return CommandResult(0, command, False)
    
    def get_player_game_square(self, player:Player) -> GameSquare:
        board_location = player.board_location
        game_square = None
        if board_location.occupation_name is not None:    # get the Occupation square
            occupation = self._careersGame.occupations[board_location.occupation_name]    # Occupation instance
            game_square = occupation.occupationSquares[board_location.occupation_square_number]
        else:       # get the border square
            game_square = self._careersGame.get_border_square(board_location.border_square_number)
        
        return game_square
    
    def get_player(self, pid, pnumber=None) -> Optional[Player] :
        """Gets a Player by initials, name, or number
            Returns: Player instance or None if no player with the given ID exists.
        """
        player = None
        if pnumber is not None:
            player = self.game_state.players[pnumber]
        else:   # lookup the player by name or initials
            players = self.game_state.players
            lc_pid = pid.lower()
            for p in players:
                if p.player_initials.lower() == lc_pid or p.player_name.lower() == lc_pid:
                    player = p
                    break
        return player
    
    #############################################################################
    #
    # command implementations. All command implementations return a CommandResult
    # Each command corresponds to a specific web server endpoint
    # 
    #############################################################################
    def roll(self, number_of_dice=2) ->CommandResult:
        """Roll 1 or 2 dice and advance that number of squares for current_player and execute the occupation or border square.
        """
        player = self.game_state.current_player
        ndice = number_of_dice
        game_square = self.get_player_game_square(player)       # could be BorderSquare or OccupationSquare

        if game_square.square_class == 'occupation':    # then I am on an occupation path so roll 1 die
            ndice = 1
        
        dice = random.choices(population=[1,2,3,4,5,6], k=ndice)
        num_spaces = sum(dice)
        
        #
        # check if the player is Unemployed and if so, if the roll allows them to move
        #
        canmove, result = self._gameEngineCommands.can_player_move(player)
        
        if canmove:
            next_square_number = self._get_next_square_number(player, num_spaces)
            
            self.log(f' {player.player_initials}  rolled {num_spaces} {dice}')
    
            #
            # place the player on the next_square_number
            #
            result = self.goto(next_square_number)
        return result
    
    def use(self, what, card_number, spaces=0) -> CommandResult:
        """Use an Experience or Opportunity card in place of rolling the die.
            Experience and Opportunity cards are identified (through the UI) by number, which uniquely
            identifies the card function. i.e. Cards having the same "number" are identical
            with respect to the end result. There are generally more than 1 card of a given number
            in the deck and this is identified by the ordinal "ncard".
            Arguments:
                what - "opportunity", "experience"
                card_number - the unique number for this card. Corresponds to card.number
                spaces - required for Experience wild cards.
        """
        player = self.game_state.current_player

        result = None
        #
        # get the actual card instance from the players deck
        #
        if what.lower() == 'opportunity':
            cards = player.get_opportunity_cards()   # dict with number as the key
            thecard_dict = cards.get(card_number, None)
            if thecard_dict is None:    # no such card
                result = CommandResult(CommandResult.ERROR, f"No {what} card exists with number {card_number} ", False)
            else:
                thecard = thecard_dict['card']
                player.opportunity_card = thecard
                result = self._execute_opportunity_card(player, opportunityCard=thecard)
        elif what.lower() == 'experience':
            cards = player.get_experience_cards()   # dict with number as the key
            thecard_dict = cards.get(card_number, None)
            if thecard is None:    # no such card
                result = CommandResult(CommandResult.ERROR, f"No {what} card exists with number {card_number} ", False)
            else:
                thecard = thecard_dict['card']
                player.experience_card = thecard
                result = self._execute_experience_card(player, experienceCard=thecard, spaces=spaces)
        else:
            result = CommandResult(CommandResult.ERROR, f"use can't use a '{what}' ", False)
            
        return result
    
    def use_insurance(self) -> CommandResult:
        """Use insurance, if you have it, to avoid paying a penalty or loosing hearts/stars/salary
            Returns: CommandMessage with return_code == SUCCESSFUL if okay, == ERROR if no insurance
                or not applicable in this case.
        """
        player = self.game_state.current_player
        if player.is_insured:
            result = CommandResult(CommandResult.SUCCESS, "'use_insurance' command not yet implemented", False)
        else:
            result = CommandResult(CommandResult.ERROR, "You don't have insurance!", False)
        return result
        
    def goto(self, square_number:int) -> CommandResult:
        """Immediately place the designated player on the designated BorderSquare OR OccupationSquare and execute that square.
            If the current player is in an occupation, this places the player on square_number of that Occupation.
            Otherwise, the square_number refers to a BorderSquare.
            NOTE that the border square number could be out of range. i.e. > game size. 
            If so, the square number adjusted and then the pass_payday() action is executed.
            
            NOTE that when in an Occupation the designated square_number could be out of range. i.e. > the occupation exit_square_number.
            If so, the player is advanced to the next BorderSquare and the exit occupation logic is executed.
            
        """
        return self._goto(square_number, self.game_state.current_player)
    
    def enter(self, occupation_name:str) -> CommandResult:
        """Enter the named occupation at the designated square number and execute the occupation square.
            This checks if the player meets the entry conditions
            and if not, return an error with the appropriate message.
            Upon entering, the player's BoardLocation  occupation_name is set to occupation_name,
            and border_square_number = current border_square_number.
            Arguments:
                occupation_name - the name of the occupation to enter. Case-sensitive!
            Return:
                CommandResult
                
            The player must either roll or use an experience card to actually enter the Occupation.
        """
        player = self.game_state.current_player
        if occupation_name in self._careersGame.occupation_names:
            occupation_entrance_square = self._careersGame.get_occupation_entrance_squares()[occupation_name]    # BorderSquare instance
            #
            # if player used an Opportunity to get here, remove that from their deck and set their opportunity_card to None
            #
            if player.opportunity_card is not None and player.opportunity_card.opportunity_type=='occupation' and player.opportunity_card == occupation_name:
                player.used_opportunity()
                    
            if self.can_enter(occupation_name, player):        # this also checks the Opportunity card used (if any)
                player.board_location.border_square_number = occupation_entrance_square.number
                player.board_location.border_square_name=occupation_entrance_square.name
                player.board_location.occupation_name=occupation_name
                player.board_location.occupation_square_number = -1     # still need to roll or use a card

                result = self.where("am","I")
                return CommandResult(result.return_code, f'{result.message}', True)
            else:
                return CommandResult(CommandResult.ERROR, "Sorry, you don't meet the occupations entry conditions.", False)
        else:
            return CommandResult(CommandResult.ERROR, "No such occupation", False)
    
    def status(self, initials:str=None) -> CommandResult:
        player = self.game_state.current_player if initials is None else self.get_player(initials)
        message = player.player_info(include_successFormula=True)
        result = CommandResult(CommandResult.SUCCESS,  message, False)
        return result       
    
    def done(self) -> CommandResult:
        """End my turn and go to the next player
        """
        cp = self.game_state.current_player
        cp.board_location.reset_prior()            # this player's prior board position no longer relevant
        npn = self.game_state.set_next_player()    # sets current_player and returns the next player number (npn) and increments turns
        player = self.game_state.current_player
        player.opportunity_card = None
        player.experience_card = None
        #
        # save the Game on change of turns
        #
        self._gameEngineCommands.save_game(self._game_filename_base, self.gameId, how='pkl')
        result = CommandResult(CommandResult.SUCCESS,  f"{cp.player_initials} Turn is complete, {player.player_initials}'s ({npn}) turn " , True)
        return result
    
    def next(self) -> CommandResult:
        """Synonym for done - go to the next player
        """
        return self.done()
    
    def end(self, save:str=None) -> CommandResult:
        """Ends the game, saves the current state if specified, and exits.
        """
        self.log("Ending game: " + self.gameId)
        if save is not None and save.lower()=='save':    # save the game state first
            sg_result = self.save_game()
            result = CommandResult(CommandResult.TERMINATE, f'Game is complete and saved to file: {sg_result.message}', True)
        else:
            result = CommandResult(CommandResult.TERMINATE, "Game is complete" , True)

        return result
    
    def quit(self, initials:str) -> CommandResult:
        """A single player, identified by initials, leaves the game.
        """
        result = CommandResult(CommandResult.SUCCESS, "'quit' command not yet implemented", False)
        return result
    
    def save(self, how="pkl") -> CommandResult:
        """Save the current game state.
            Arguments: how - save format: 'json' or 'pkl' (the default).
            save('pkl') uses joblib.dump() to save the CareersGame instance to binary pickel format.
            This can be reconstituted with joblib.load()
        """
        return self._gameEngineCommands.save_game(self._game_filename_base, self.gameId, how=how)
    
    def where(self, t1:str="am", t2:str="I") -> CommandResult:
        """where am I or where is <player>
        """
        player = None 
        
        result = CommandResult.SUCCESS
        if t1=='am' and t2.lower() == 'i':
            player = self.game_state.current_player
            message = "You are on square# "
        elif t1=='is':
            # t2 is a player's initials or name
            message = f'{t2} is on square# '
            player = self.get_player(t2)
        if player is not None:
            current_board_location = player.board_location
            game_square = self.get_player_game_square(player)
            message += str(game_square.number)
            if game_square.square_class == 'Occupation':
                message += " of " + current_board_location.occupation_name + ": '" +  game_square.text + "' " 
            else:    # a border square
                message += ": " + game_square.name
            if game_square.action_text is not None:
                message += "\n" + game_square.action_text
        else:
            message = f'No such player: {t2}'
            result = CommandResult.ERROR
        
        return CommandResult(result, message, False)
    
    def retire(self) -> CommandResult:
        """Retire this player to the retirement corner square (Spring Break, Holiday)
        
        """
        result = CommandResult(CommandResult.SUCCESS, "'retire' command not yet implemented", False)
        return result
    
    def bump(self, who) -> CommandResult:
        """The current player bumps another player occupying the same board square to Unemployment
            Note that it is possible to land on a square occupied by more than one player,
            for example if a player chooses NOT to bump that square will have 2 players.
            In any case the initials of the player to be bumped must be provided.
            Only a single player may be bumped even if there are more than one on the square.
        
        """
        player = self.game_state.current_player
        bumped_player = None
        for p in player.can_bump:
            if p.player_initials == who:
                bumped_player = p
        
        if bumped_player is None:
            result = CommandResult(CommandResult.ERROR, f'Player {who} cannot be Bumped')
        else:
            #
            # find the Unemployment square and place the bumped_player there
            # and also set the is_unemployed flag
            #
            border_square = self._careersGame.find_border_square("Unemployment")
            result = border_square.execute(bumped_player)
        return result
    
    def bankrupt(self) -> CommandResult:
        """The current player declares bankruptcy.
            A bankrupt player looses all cash, experience and opportunity cards and essentially
            restarts the game with configured starting cash and salary, and positioned at the "Payday" square (border square 0).
            The player does retain occupation experience however including any and all college degrees.
        
        """
        result = CommandResult(CommandResult.SUCCESS, "'bankrupt' command not yet implemented", False)
        return result
    
    def pay(self, amount_str, initials=None) -> CommandResult:
        """The current player, or the player whose initials are provided, makes a payment associated with their current board position.
            If the paying player has sufficient cash to make the payment that amount is subtracted
            from their cash on hand and CommandResult.SUCCESS with done_flag = True is returned 
            Otherwise, CommandResult.ERROR is returned with done_flag = False.
            This does NOT bankrupt the player automatically because a player may bargain with other
            player(s) to get a loan to pay off the debt.
            If no such arrangement is made (in the UI), a bankrupt command is sent.
            Experience and Opportunity cards can be transfered from one player to another using the transfer() command.
            
            NOTE - "loans" are not paid off automatically. HOWEVER a player's net worth, which determines total_points,
            will subtract all loan amounts from cash on hand.
            
        """
        amount = int(amount_str)
        player = self.game_state.current_player if initials is None else self.get_player(initials)
        if player.cash >= amount:
            player.cash = player.cash - amount
            message = f'{player.player_initials} paid {amount}'
            return CommandResult(CommandResult.SUCCESS, message, True)
        else:
            message = f'{player.player_initials} has insufficient funds to cover {amount} and must either borrow cash from another player or declare bankruptcy'
            return CommandResult(CommandResult.ERROR, message, False)
        
    def transfer(self, quantity:int, what:str, from_player_number:int) -> CommandResult:
        """Transfers cash, opportunities or experience cards from one player to the current player
            Arguments:
                from_player_number - the player number transferring
                what - "cash", "opportunity", "experience"
                quantity - the amount of cash or number of experience/opportunity cards being transferred
        """
        CommandResult(CommandResult.SUCCESS, "transfer command not implemented", True)
        
    
    def list(self, what='all') ->CommandResult:
        """List the Experience or Opportunity cards held by the current player
            Arguments: what - 'experience', 'opportunity', or 'all'
            Returns: CommandResult.message is the stringified list of str(card).
                For Opportunity cards this is the text property.
                For Experience cards this is the number of spaces (if type is fixed), otherwise the type.
            
        """
        player = self.game_state.current_player
        return GameEngineCommands.list(player, what)    

    def saved(self) -> CommandResult:
        """List the games saved by this master_id, if any
        
        """
        result = CommandResult(CommandResult.SUCCESS, "'saved' command not yet implemented", False)
        return result    

    def load(self, gameid:str) -> CommandResult:
        """Load a previously saved game, identified by the game Id
        
        """
        result = CommandResult(CommandResult.SUCCESS, "'load' command not yet implemented", False)
        return result    

    def who(self, t1:str="am", t2:str="I") -> CommandResult:
        """Who am I or who is <player>
            Returns: A CommandResult where message is the player's info: name, initials, number, SSN (kidding)
                If there is no player with initials t2, return_code is ERROR.
        """
        result = CommandResult.SUCCESS
        if t1=='am' and t2.lower() == 'i':
            player = self.game_state.current_player
            message = player.info()
        elif t1=='is':
            # t2 is a player's initials or name
            player = self.get_player(t2)
            if player is not None:
                message = player.info()
            else:
                message = f'No such player: {t2} '
                result = CommandResult.ERROR
            
        return CommandResult(result, message, True)
    
    def add(self, what, name, initials=None, stars=0, hearts=0, cash=0) -> CommandResult:
        """Add a new player to the Game OR add a degree to the current player or the player whose initials are provided.
    
        """
        if what == 'player':
            sf = SuccessFormula(stars=stars, hearts=hearts, cash=cash)
            player = Player(name=name, initials=initials)
            player.success_formula = sf
            player.salary = self._careersGame.game_parameters['starting_salary']
            player.cash = self._careersGame.game_parameters['starting_cash']
            self._careersGame.add_player(player)        # adds to GameState
            
            message = f'Player "{name}" "{initials} number: {player.number}" added'
        else:    
            # adds a degree in the current (or named) player's degree programs
            # name is the name of the degree program
            player = self.game_state.current_player if initials is None else self.get_player(initials)
            result = self.add_degree(player, name)
            message = result.message
        
        self.log(message)
        return CommandResult(CommandResult.SUCCESS, message, False)
    
    def game_status(self) -> CommandResult:
        """Get information about the current game in progress and return in JSON format
        """
        message =  self.game_state.to_JSON()
        
        return CommandResult(CommandResult.SUCCESS, message, True)
    
    def create(self, edition, master_id, game_type, points, game_id=None) -> CommandResult:
        """Create a new CareersGame.
        
        """
        assert master_id is not None and len(master_id) >= 5
        self._edition = edition    # 'Hi-Tech'
        self._master_id = master_id
        #
        # Create the CareersGame instance and the GameEngineCommands
        #
        self._careersGame = CareersGame(self._edition, master_id, points, game_id, game_type=game_type)
        
        self._game_state = self._careersGame.game_state
        self._gameId = self._careersGame.gameId
        self._logfile_filename = "careers_" + self._careersGame.edition_name
        self._logfile_folder = "/data/log"    # TODO put in Environment
        self._gamefile_folder = "/data/games"
        self._logfile_path = self._logfile_folder + "/" + self._logfile_filename + "_" + self._gameId + ".log"
        self._game_filename_base = f'{self._gamefile_folder}/{self.master_id}_{self.gameId}_game'
        
        self.fp = open(self._logfile_path, "w")   # log file open channel
        self._gameEngineCommands = GameEngineCommands(self._careersGame, self.fp)
        self._gameEngineCommands.trace = self.trace

        message = f'Created game {self._gameId} for {self._master_id}'
        self.log(message)
        return CommandResult(CommandResult.SUCCESS, message, True)
    
    def start(self) -> CommandResult:
        return self._start()
    
    def buy(self, what, qty_str, amount_str) -> CommandResult:
        """Buy a number of items for the current player
            Arguments:
                what - "hearts" | "stars" | "experience" | "opportunity"
                qty - how many to buy (adds to the player's score or card deck)
                amount - cash amount cost
            Returns:
                CommandResult.SUCCESS if the player can cover the cost, otherwise CommandResult.ERROR with an appropriate error message
            Okay to have a negative quantity and zero amount. For example, to lose 1 heart send: "buy hearts -1 0"
        """
        return self._buy(what, qty_str, amount_str)
    
    #####################################
    #
    # Game engine action implementations
    #
    #####################################
        
    def _start(self) -> CommandResult:
        message = f'Starting game {self.gameId}'

        self.log(message)
        self.game_state.set_next_player()    # sets the player number to 0 and the curent_player Player reference
        return CommandResult(CommandResult.SUCCESS, message, True)
    
    def _buy(self, what, qty_str, amount_str) -> CommandResult:
        """Implements the buy command
        """
        player = self.game_state.current_player
        amount = int(amount_str)
        qty = int(qty_str)
        if player.cash < amount:
            return CommandResult(CommandResult.ERROR, f'Insufficient funds {player.cash} for amount {amount}', True)
        player.add_cash(amount)
        if what.lower().startswith('heart'):
            player.add_hearts(qty)
        elif what.lower().startswith('star'):
            player.add_stars(qty)
        elif what.lower().startswith('exp'):
            self.add_experience_cards(player, qty)
        elif what.lower().startswith('opp'):
            self.add_opportunity_cards(player, qty)
        else:
            return CommandResult(CommandResult.ERROR, f'Cannot add {qty} {what}', False)
        return CommandResult(CommandResult.SUCCESS, f'{qty} {what} added', True)

    def can_enter(self, occupation_name, player:Player):
        """Determine if this player can enter the named Occupation
            This checks if the player meets the entry conditions, namely:
                * it's college and they have the tuition amount in cash
                * they've previously completed this occupation and can therefore enter for free
                * or they have a qualifying degree and can therefore enter for free
                * or they have executed an "All expenses paid" Opportunity card
                * or they have sufficient cash to cover the entry fee 
        """

        occupation = self._careersGame.occupations[occupation_name]   # Occupation instance
        entry_fee = occupation.entryFee
        has_fee = player.cash >= entry_fee
        occupationClass = occupation.occupationClass
        # anyone can go to college if they have the funds
        if occupationClass == 'college' and has_fee:
            return True    
        
        #
        # check occupation record for prior trips through 
        #
        if occupation_name in player.occupation_record and player.occupation_record[occupation_name] > 0:
            return True
        #
        # check degree requirements
        #
        degreeRequirements = occupation.degreeRequirements
        degreeName = degreeRequirements['degreeName']
        numberRequired = degreeRequirements['numberRequired']
        if degreeName in player.my_degrees and player.my_degrees[degreeName] >= numberRequired:
            return True
        #
        # is the player using a  "All expenses paid" Opportunity card ?
        #
        if player.opportunity_card is not None:
            if player.opportunity_card.expenses_paid:
                return True
        
        return has_fee

    
    def _execute_opportunity_card(self,  player:Player, opportunityCard:OpportunityCard=None) -> CommandResult:
            player.opportunity_card = opportunityCard
            message = f'{player.player_initials} Playing: {opportunityCard.text}'
            #
            # Now execute this Opportunity card
            #
            self.log(message)
            result = CommandResult(CommandResult.SUCCESS, message, False)   #  TODO
            return result                   
    
    def _execute_experience_card(self,  player:Player, experienceCard:ExperienceCard, spaces=0) -> CommandResult:
            player.experience_card = experienceCard
            nspaces = experienceCard.spaces
            if experienceCard.spaces == 0 and experienceCard.type != 'fixed':    # must be a wild card
                nspaces = spaces

            message = f'{player.player_initials} Moving: {nspaces} spaces'
            self.log(message)
            result = CommandResult(CommandResult.SUCCESS, message, False)   #  TODO
            return result        
    
    def execute_game_square(self, player:Player, board_location:BoardLocation) -> CommandResult:
        """Executes the actions associated with a given BoardLocation and Player
        
        """
        game_square = self.get_player_game_square(player)
        if board_location.occupation_name is not None:
            result = self._execute_occupation_square(player, game_square, board_location)
        else:
            result = self._execute_border_square(player, game_square, board_location)
        
        return result
        
    def _execute_occupation_square(self, player:Player, game_square:OccupationSquare, board_location:BoardLocation) -> CommandResult:
        message = f'execute_occupation_square {board_location.occupation_name}  {board_location.occupation_square_number} for {player.player_initials}'
        self.log(message)
        
        return CommandResult(CommandResult.SUCCESS, message, False)
    
    def _execute_border_square(self, player:Player, game_square:BorderSquare, board_location:BoardLocation) -> CommandResult:
        message = f'{player.player_initials} landed on {game_square.name}  ({board_location.border_square_number})'
        self.log(message)
        result = game_square.execute(player)
        result.message = message + "\n" + result.message
        #
        # if there is a next action to perform then do it
        #
        if result.next_action is not None:
            action_result = self.execute_command(result.next_action, player)
            result.message = result.message + "\n" + action_result.message
            result.return_code = action_result.return_code
            result.done_flag = action_result.done_flag
        return result
    
    def _get_next_square_number(self, player:Player, num_spaces:int) -> int:
        """Gets the next square number given the number of spaces to advance.
            Arguments:
                player - the Player
                num_spaces - the number of spaces to advance. Must be >0 to go anywhere, but this doesn't check.
            Returns: the next square number, depending on whether the player is currently on a BorderSquare or OccupationSquare
            NOTE that the square number returned could be out of bounds for the occupation or along the border.
            This is handled by goto() which does the actual placement.
        """
        board_location = player.board_location
        game_square = self.get_player_game_square(player)       # could be BorderSquare or OccupationSquare

        #
        # advance that number of spaces if permitted to do so. 
        # If the player is on the occupation entrance square, the occupation_square_number will be -1
        # Squares are numbered starting with 0
        # 
        if game_square.square_class == 'occupation':
            next_square_number = board_location.occupation_square_number + num_spaces    # could be > size of the occupation, that's handled by goto()
        else:
            next_square_number = board_location.border_square_number + num_spaces        # could be > size of the board, that's also handled by goto()
            
        return next_square_number
    

    def _goto(self, square_number:int, player:Player) -> CommandResult:
        """Immediately place the designated player on the designated BorderSquare OR OccupationSquare and execute that square.
            If the current player is in an occupation, this places the player on square_number of that Occupation.
            Otherwise, the square_number refers to a BorderSquare.
            NOTE that the border square number could be out of range. i.e. > game size. 
            If so, the square number adjusted and then the pass_payday() action is executed.
            
            NOTE that when in an Occupation the designated square_number could be out of range. i.e. > the occupation exit_square_number.
            If so, the player is advanced to the next BorderSquare and the exit occupation logic is executed.
            
        """
        game_square = self.get_player_game_square(player)   # where player currently is on the board
        board_location = player.board_location
        result = None
        if game_square.square_class == 'occupation':        # If player is currently in or completing an Occupation
            
            occupation_name =  player.board_location.occupation_name
            occupation = self._careersGame.occupations[occupation_name]
            if square_number >= occupation.size:
                
                exit_result =  self.exit_occupation(player, board_location)
                
                #
                # go to next border square and update board_location after the occupation exit
                #
                board_location.border_square_number = square_number - occupation.size + occupation.exit_square_number - 1
                board_location.occupation_name = None
                game_square =  self._careersGame.get_border_square(board_location.border_square_number)
                board_location.border_square_name = game_square.name                
                
                result = self.execute_game_square(player, board_location)
                result.message = exit_result.message + "\n" + result.message
                #return result
            else:    # next square is in this occupation
                board_location.occupation_square_number = square_number
                return self.execute_game_square(player, board_location)
                
        else:   # goto designated border square
            if square_number >= 0 and \
               square_number < self._careersGame.game_board.game_layout_dimensions['size'] and \
               square_number > board_location.border_square_number :
                player = self.game_state.current_player
                border_square = self._careersGame.get_border_square(square_number)
                board_location.border_square_number = square_number
                board_location.border_square_name = border_square.name
                board_location.occupation_name = None
                #
                # execute this game square
                # unless it's a travel_square and we just game from a travel_square
                # otherwise we'd get into an endless travel loop
                #
                if border_square.square_type == 'travel_square' and self.was_prior_travel(player):
                    result = CommandResult(CommandResult.SUCCESS, "", True)
                else:
                    result = self.execute_game_square(player, board_location)
                #return result
            else:    # player passed or landed on Payday
                if square_number >  board_location.border_square_number :
                    square_number = square_number - self._careersGame.game_board.game_board_size
                border_square = self._careersGame.get_border_square(square_number)
                board_location.border_square_number = square_number
                board_location.border_square_name = border_square.name
                payday_result = self.pass_payday(player)
                #
                # again, check travel arrangements
                #
                if border_square.square_type == 'travel_square' and self.was_prior_travel(player):
                    result = CommandResult(CommandResult.SUCCESS, "", True)
                else:
                    result = self.execute_game_square(player, board_location)
            
                result.message = payday_result.message + "\n" + result.message
                #return result
        #
        # if another player is on that square AND the square is NOT Unemployment, they can be bumped
        #
        other_players = self.who_occupies_my_square(player)
        if player.current_border_square_name() != "Unemployment":
            if len(other_players) > 0:
                initials = [p.player_initials for p in other_players]
                result.message = result.message + f'\n{player.player_initials} can bump {initials}'
        player.can_bump = other_players   # could be an empty list
        return result

    def exit_occupation(self, player, board_location:BoardLocation) -> CommandResult:
        """Applies exiting an occupation rules when a player exits an occupation path.
            Exiting can be done by rolling out, using an Experience, or using an Opportunity to go somewhere else.
            The occupation could be College ("occupationClass" : "college") or a regular Occupation ("occupationClass" : "occupation")
            * if the player entered the Occupation via Opportunity card, that card is removed from the player's hand
            * if the player completed the Occupation, credit is applied to the player's occupation record and up to 3 Experience cards given
            
            Note that exit_occupation does NOT update the player's board position.  
            That is done by goto() which calls this method before updating the board position.
            So upon entry, border square location is the occupation entrance square for this occupation (or college).
            Note also that the player's salary is NOT adjusted when completing college. That is done by the add_degree() method.
            
        """
        nexperience = 0
        trips = 1
        if board_location.occupation_name in player.occupation_record:
            trips = player.occupation_record[board_location.occupation_name]   # number of trips not counting this one
            player.occupation_record[board_location.occupation_name] = trips + 1
            
        else:
            player.occupation_record[board_location.occupation_name] = trips
        
        #
        # No experience given for completing College
        # 
        game_square = self._careersGame.get_border_square(board_location.border_square_number)
        if game_square.square_class == 'occupation_entrance_square':
            if trips <= 3:  # at most 3 experience cards can be given
                nexperience = trips
            else:
                nexperience = 3

        self.add_experience_cards(player, nexperience)
        
        message = f'{player.initials} Leaving {board_location.occupation_name}, collects {nexperience} Experience cards'
        self.log(message)
        result = CommandResult(CommandResult.SUCCESS, message, True)
        return result
    
    def add_experience_cards(self, player:Player, ncards:int):
        deck = self._careersGame.experience_cards     # ExperienceCardDeck
        for i in range(ncards):
            card = deck.draw()
            player.add_experience_card(card)
            
    def add_opportunity_cards(self, player:Player, ncards:int):
        deck = self._careersGame.opportunities        # OpportunityCardDeck
        for i in range(ncards):
            card = deck.draw()
            player.add_opportunity_card(card)
        
    def pass_payday(self, player:Player) -> CommandResult:
        """Performs any actions associated with landing on or passing the Payday square.
            Count laps (which may or may not be used) and dole out appropriate salary.
            Arguments:
                player - current Player reference
            The player's board location reflects the new position which will have a border_square_number >= 0.
        """
        game_square = self._careersGame.get_border_square(0)    # Payday is always square 0
        result = game_square.execute(player)
        self.log(result.message)
        return result
    
    def was_prior_travel(self, player) -> bool:
        """Returns True if the player's prior border position was a travel square, False otherwise
            This is needed in order to avoid endless travel hops.
        """
        result = False
        ps = player.board_location.prior_border_square_number
        if ps is not None and self._careersGame.game_board.border_squares[ps].square_type == 'travel_square':
            result = True
        return result
        
    def add_degree(self, player, degreeProgram) -> CommandResult:
        """Adds a degree to the player and adjusts the salary as needed.
            The maximum number of degrees a player can have in any degree program is "maxDegrees".
            Player's Salary is not adjusted if their number of degrees exceeds that.
        """
        game_degrees = self._careersGame.college_degrees
        degree_names = game_degrees['degreeNames']
        if degreeProgram not in game_degrees['degreePrograms']:
            return CommandResult(CommandResult.ERROR, f"No such degree program: {degreeProgram}", False)
        
        max_allowed = game_degrees['maxDegrees']
        salary_increases = game_degrees['salaryIncreases']
        my_degrees = player.my_degrees
        ndegrees = 0
        salary_inc = 0
        if degreeProgram in my_degrees:
            ndegrees = player.my_degrees[degreeProgram]
        
        if ndegrees == 0:
            player.my_degrees[degreeProgram] = 1
            salary_inc = salary_increases[0]
            player.salary = player.salary + salary_inc   # automatically adds to salary_history
            
        elif ndegrees + 1 <= max_allowed:
            salary_inc = salary_increases[ndegrees]
            player.my_degrees[degreeProgram] = ndegrees + 1
            player.salary = player.salary + salary_inc
        
        else:
            ndegrees = 3  # already maxed out
        
        message = f'{player.player_initials} awarded {degree_names[ndegrees]} degree in {degreeProgram}'
        if salary_inc > 0:
            currency = self._careersGame.game_parameters['currency']
            message += f' and a Salary increase of {salary_inc} {currency}'
        
        return CommandResult(CommandResult.SUCCESS, message, True)
        
    def who_occupies_my_square(self, player) -> List[Player]:
        """If the game square currently occupied by player is also occupied by another player(s), return those players.
            The player landing on an occupied square may bump that player to Unemployment.
            Arguments: 
                player - the Player landing on the designated square
            Returns: a list of Player on the square occupied by player, or an empty list if none.
        """
        plist = []
        occupation_name = player.current_occupation_name()      # could be None
        for ap in self._game_state.players:
            if player.number == ap.number:
                continue    # it's me!
            if occupation_name is None:     # player is on a border square
                if player.current_border_square_number() == ap.current_border_square_number():
                    plist.append(ap)
            else:   # player is in an occupation
                if ap.current_occupation_name() == occupation_name and ap.current_occupation_square_number() == ap.current_occupation_square_number():
                    plist.append(ap)
                
        return plist
    
        


import uvicorn
from fastapi import FastAPI, Depends
from server.gameManager import CareersGameManager

manager = CareersGameManager()    
app = FastAPI()

@app.get("/", status_code=200)
def get(game: CareersGameManager=Depends(manager)):
    return {}

if __name__ == "__main__":
    uvicorn.run("app:app", port=9000, reload=True)


'''
Created on Aug 9, 2022

@author: don_bacon
'''
import json
from pathlib import Path
from game.gameUtils import GameUtils

class CardDeck(object):
    """Abstract class representing a deck of game cards a player draws or is given.
    
    """


    def __init__(self, resource_path, deck_name, edition_name):
        '''
        Constructor
        '''
        self._resource_path = resource_path
        self._deck_name = deck_name         # "opportunityCards" or "experienceCards" for example
        self._edition_name = edition_name
        self._cards_content = self.load_cards(resource_path, deck_name, edition_name)
        self._type_list = self._cards_content['type_list']
        self._help = self._cards_content['Help']
        self._card_types = self._cards_content['types']
        self._cards = self._cards_content['cards']
        
        self._deck = []

        self._cards_index = []
        self._size = self.create_card_deck()
        self._cards_index = GameUtils.shuffle(self._size)
        self._next_index = 0
        
    def shuffle(self):
        self._cards_index = GameUtils.shuffle(self._size)
        
    @property
    def size(self):
        return self._size
    
    @property
    def next_index(self):
        return self._next_index
    
    @next_index.setter
    def next_index(self, value):
        self._next_index = value
    
    def draw(self):
        """Draw a card from the deck. If no cards remaining, re-shuffle and reset the next_index
        
        """
        next_card = None
        if self.next_index < self.size:    # draw the next card
            nxt = self._cards_index[self.next_index]
            next_card = self._deck[nxt]
            self.next_index = self.next_index + 1
        else:
            self.next_index = 0
            self.shuffle()
            return self.draw()
        return next_card
    
    def load_cards(self, path, name, edition_name):
        """Loads the specified card deck JSON file. This assumes the resulting structure is a dict
            with keys: "Help", "type_list", "types", and "cards"
            Arguments:
                path - JSON file path
                name - the card deck name, for example "opportunityCards"
                edition_name
            Returns:
                the JSON content as a dict
        """
        cards = dict()
        filepath = f'{path}/{name}_{edition_name}.json'
        p = Path(filepath)
        if p.exists():
            fp = open(filepath, "r")
            cards = json.loads(fp.read())
        return cards
    
    def create_card_deck(self):
        count = 0
        for card_spec in self._cards:
            qty = card_spec['quantity']
            count += qty
            self.save_card(card_spec, qty)
        return count
        
    def save_card(self, card_spec, qty):
        """Saves a single Opportunity or Experience card.
        
        Abstract method - Override in concrete class.
        """
        pass
    
    
    


'''
Created on Aug 20, 2022

@author: don_bacon
'''
from careers.game.occupationSquare import OccupationSquare

class Occupation(object):
    """Encapsulates a Careers Occupation.
    
    """


    def __init__(self, occupation_dict:dict, game=None):
        """
        Constructor
        """
        self._name = occupation_dict['name']
        self._configuration = occupation_dict['configuration']
        self._occupationClass = self._configuration['occupationClass']
        self._entryFee = self._configuration['entryFee']
        self._size = self._configuration['size']
        self._entry_square_number = self._configuration['entrance_square_number']
        self._exit_square_number = self._configuration['exit_square_number']
        self._entry_text = self._configuration['text']
        self._fullName = self._configuration['fullName']
        self._degreeRequirements = self._configuration.get('degreeRequirements', None)
        self._occupationSquares = self._create_occupation_squares(occupation_dict["occupationSquares"], game)
        self._double_happiness = False      # this is temporarily set by a "double_happiness" Opportunity card
        
    
    def _create_occupation_squares(self, occupationSquares:list, game) -> list:
        """For this Occupation create a list of OccupationSquare corresponding to the "occupationSquares".
            Returns: a List of  OccupationSquare
        """
        occupation_squares = list()
        for occupation_square_dict in occupationSquares:
            occupation_square = OccupationSquare(occupation_square_dict, game=game)
            occupation_squares.append(occupation_square)
        return occupation_squares
    
    @property
    def name(self):
        return self._name
    
    @property
    def occupationClass(self):
        return self._occupationClass
    
    @property
    def entryFee(self):
        return self._entryFee
    
    @property
    def size(self):
        return self._size
    
    @property
    def entry_square_number(self):
        return self._entry_square_number
    
    @property
    def exit_square_number(self):
        return self._exit_square_number
    
    @property
    def fullName(self):
        return self._fullName
    
    @property
    def occupationSquares(self):
        return self._occupationSquares
    
    @property
    def degreeRequirements(self):
        return self._degreeRequirements
    
    @property
    def double_happiness(self):
        return self._double_happiness
    
    @double_happiness.setter
    def double_happiness(self, value):
        self._double_happiness = value
        
        


'''
Created on Aug 15, 2022

@author: don_bacon
'''
from datetime import datetime
import json
from game.player import Player
from game.careersObject import CareersObject

class GameState(CareersObject):
    """Maintains the global state of a Careers game instance.
    
    """


    def __init__(self, total_points, game_type='points'):
        """Create and initialize a Careers GameState
            The GameState encapsulates the following dynamic game state properties:
                number_of_players
                players - a [Player]
                current_player_number -  the current player number (whose turn it is)
                current_player - a Player instance
                winning_player - a Player instance, None until the game determines a winner
                total_points - points needed to win if game_type is 'points', else it's the #minutes in a timed game
                turns - total number of completed turns taken by all players
                turn_number - the current turn number
                game_start - the datetime when the game started (when this object was created)
                seconds_remaining - number of seconds remaining if a timed game
                game_type - 'points' (the default) or 'timed'
        """
        self._number_of_players = 0
        self._players = []   # list of Player
        self._current_player_number = -1
        self._current_player = None
        self._total_points = total_points
        self._winning_player = None
        self._turns = 0
        self._turn_number = 1
        self._game_start = datetime.now()
        self._game_complete = False
        
        self._game_type = game_type
        self._seconds_remaining = 0 if game_type=='points' else total_points * 60
    
    @property
    def number_of_players(self):
        return self._number_of_players
    
    @number_of_players.setter
    def number_of_players(self, value):
        self._number_of_players = value

    @property
    def players(self):
        return self._players
    
    @property
    def winning_player(self):
        return self._winning_player
    
    @property
    def current_player(self):
        return self._current_player
    
    @current_player.setter
    def current_player(self, value):
        self._current_player = value
    
    @property
    def current_player_number(self):
        return self._current_player_number
    
    @current_player_number.setter
    def current_player_number(self, value):
        self._current_player_number = value
        
    @property
    def game_type(self):
        return self._game_type
    
    @property
    def game_complete(self):
        return self._game_complete
    
    @property
    def total_points(self):
        return self._total_points
    
    @property
    def turns(self):
        return self._turns
    
    @turns.setter
    def turns(self, value):
        self._turns = value
    
    @property
    def turn_number(self):
        return self._turn_number
    
    @turn_number.setter
    def turn_number(self, value):
        self._turn_number = value
    
    @property
    def seconds_remaining(self):
        return self._seconds_remaining

    def set_next_player(self):
        """Returns the player number of the next player. And sets the value of current_player.
            If the next player has lose_turn == True, the current_player is set to the next player
            after that player and their lose_turn flag is reset to False.
        
        """
        npn = self._get_next_player_number()
        if self.players[npn].lose_turn:
            self.current_player_number = npn
            self.players[npn].lose_turn = False
            npn = self._get_next_player_number()
            
        self.current_player_number = npn    
        self.current_player = self.players[self.current_player_number]
        if npn == 0:
            self.turns += 1
            self.turn_number += 1
            
        return self.current_player_number
    
    def _get_next_player_number(self):
        p = self.current_player_number + 1
        if p >= self.number_of_players:
            return 0
        else:
            return p
    
    def add_player(self, aplayer:Player):
        """Add a Player to the game and increments the number_of_players.
        
        """
        aplayer.number = self.number_of_players     # starts at 0
        self._players.append(aplayer)
        self._number_of_players += 1
    
    def get_player_by_initials(self, initials):
        player = None
        for p in self.players:
            if p.player_initials.lower() == initials.lower():
                player = p
                break
        return player
    
    def to_JSON(self):
        
        gs = {"game_type" : self.game_type, "number_of_players" : self.number_of_players, "current_player_number" : self.current_player_number }
        gs["turns"] = self.turns
        gs["turn_number"] = self.turn_number
        gs["total_points"] = self.total_points
        gs["seconds_remaining"] = self.seconds_remaining
        if self.winning_player is not None:
            gs["winning_player"] = self.winning_player.initials
        gs["game_complete"] = self.game_complete
        #gs["players"] 

        players = []
        for player in self.players:
            players.append(player.to_dict())
        gs["players"] = players

        return json.dumps(gs, indent=2)
        


'''
Created on Aug 24, 2022

@author: don_bacon
'''
from game.careersObject import CareersObject
import json

class BoardLocation(CareersObject):
    """Defines a position on the game board.
        Board location is defined by 4 properties:
            border_square_number - the border square number which is in the range(0, game_board_size +1)
            occupation_name - the Occupation name IF the player is currently on an occupation path, else None
            occupation_square_number - if occupation_name is not None, the square number of that occupation
            border_square_name - (optional) if not None the name of this border square (from the game layout)
            json_string - JSON representation, optional.
        
        NOTE that if the player is on an occupation path, the border_square_number will be
        the border square number of that occupation's entrance square.
    
    """

    def __init__(self, border_square_number=0, border_square_name=None, occupation_name=None, occupation_square_number=0, json_string=None):
        """
        Constructor
        """
        if json_string is not None:
            self.from_JSON(json_string)
        else:
            self._border_square_number = border_square_number
            self._border_square_name = border_square_name
            self._occupation_name = occupation_name
            self._occupation_square_number = occupation_square_number
        #
        # these properties track where we came from and are all read-only
        #
        self._prior_border_square_number = None
        self._prior_border_square_name = None
        self._prior_occupation_square_number = None
        self._prior_occupation_name = None
    
    def from_JSON(self, jstr):
        bs_dict = json.loads(jstr)    # this could raise a TypeError
        self.border_square_number = bs_dict['border_square_number']
        self.border_square_name = bs_dict['border_square_name']
        self.occupation_name = bs_dict['occupation_name']
        self.occupation_square_number = bs_dict['occupation_square_number']
    
    @property
    def border_square_number(self):
        return self._border_square_number
    
    @border_square_number.setter
    def border_square_number(self, value):
        self._prior_border_square_number = self.border_square_number
        self._border_square_number = value
        
    @property
    def occupation_name(self):
        return self._occupation_name
    
    @occupation_name.setter
    def occupation_name(self, value):
        self._prior_occupation_name = self.occupation_name
        self._occupation_name = value
        
    @property
    def occupation_square_number(self):
        return self._occupation_square_number
    
    @occupation_square_number.setter
    def occupation_square_number(self, value):
        self._prior_occupation_square_number = self.occupation_square_number
        self._occupation_square_number = value
        
    @property
    def border_square_name(self):
        return self._border_square_name
    
    @border_square_name.setter
    def border_square_name(self, value):
        self._prior_border_square_name = self.border_square_name
        self._border_square_name = value
        
    @property
    def prior_border_square_number(self):
        return self._prior_border_square_number
    @property
    def prior_border_square_name(self):
        return self._prior_border_square_name
    @property
    def prior_occupation_square_number(self):
        return self._occupation_square_number
    def prior_occupation_name(self):
        return self._prior_occupation_name
    
    def reset_prior(self):
        """After a player's turn is complete their prior location is no longer relevant. This sets all the prior_ values to None
        """
        self._prior_border_square_number = None
        self._prior_border_square_name = None
        self._prior_occupation_name = None
        self._prior_occupation_square_number = None
        
    def __str__(self):
        locn = f'Border square#: {self.border_square_number}'
        if self.border_square_name is not None:
            locn += f'  "{self.border_square_name}"'
        if self.occupation_name is not None:
            locn += f'\tOccupation: {self.occupation_name}   square#: {self._occupation_square_number}'
        return locn
    
    def to_dict(self):
        return {"border_square_number":self.border_square_number, "border_square_name":self.border_square_name, \
                "occupation_square_number":self.occupation_square_number, "occupation_name":self.occupation_name }
    
    def to_JSON(self):
        return json.dumps(self.to_dict(), indent=2)
    



from setuptools import find_packages, setup
import os

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

#
# the README file content is the long description
#
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

requirements = [
]

setup(
    name="careers",
    version="0.0.6",
    author="Donald Bacon",
    author_email="dwbzen@gmail.com",
    description="Careers Game game engine and server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dwbzen/careers.git",
    packages=find_packages(exclude=["test",]),
    data_files=[('resources', ['resources/editions.json','resources/experienceCards_Hi-Tech.json','resources/opportunityCards_Hi-Tech.json',
                'resources/collegeDegrees_Hi-Tech.json', 'resources/gameLayout_Hi-Tech.json', 'resources/occupations_Hi-Tech.json',
                'resources/FMC_Hi-Tech.json','resources/UF_Hi-Tech.json', 'resources/Amazon_Hi-Tech.json', 'resources/ESPN_Hi-Tech.json',
                'resources/Disney_Hi-Tech.json','resources/Facebook_Hi-Tech.json','resources/Google_Hi-Tech.json',
                'resources/ListerAndBacon_Hi-Tech.json','resources/Pfizer_Hi-Tech.json', 'resources/SpaceX_Hi-Tech.json'
                ])],
    install_requires=requirements,
    license="MIT",
    namespace_packages=[],
    include_package_data=False,
    zip_safe=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
		"Intended Audience :: Developers",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
	    "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires='>=3.8',
    keywords="careers game",
    maintainer='Donald Bacon',
    maintainer_email='dwbzen@gmail.com'
)



"""Not sure what this is going to do yet.
    Probably managed unique GameEngine instances (one per game).
"""

from typing import Any

class CareersGameManager(object):

    def __init__(self):
        pass

    def __call__(self) -> Any:
        return self


'''
Created on Aug 06, 2022

@author: don_bacon
'''

VERSION_TUPLE = (0, 0, 6)
__version__ = ".".join(map(str, VERSION_TUPLE))
__title__ = 'server'
__description__ = 'Python Careers game'
__author__ = 'Donald Bacon'
__author_email__ = 'dwbzen@gmail.com'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2022 Donald Bacon'


'''
Created on Aug 9, 2022

@author: don_bacon
'''

from game.cardDeck import CardDeck
from game.opportunityCard import OpportunityCard

class OpportunityCardDeck(CardDeck):
    """The deck of Opportunity Cards used in game play.
    
    """

    def __init__(self, resource_path, edition_name):
        '''
        Constructor
        '''
        super().__init__(resource_path, "opportunityCards", edition_name)    # loads the card deck
        self._opportunity_types = ["occupation", "occupation_choice", "border_square", "border_square_choice", "action", "travel", "opportunity" ]

        
    def save_card(self, card_spec:dict, qty):
        """Appends n-instances of a single Opportunity card to the card deck.
            Arguments: card_spec - the JSON opportunity card content as a dict
            
        card_spec includes:
            "number" - the unique card number. Cards having the same number are identical.
            "quantity" - specifies the number of this card that appear in the deck.
        
        """
        number = card_spec['number']
        destination = card_spec['destination']
        ctype = card_spec['type']
        expenses_paid = card_spec.get('expenses_paid', 0) == 1
        double_happiness = card_spec.get('double_happiness', 0) == 1

        for ncard in range(1, qty+1):
            opportunity_card = OpportunityCard(ctype, number, ncard, destination, card_spec['text'], expenses_paid, double_happiness)
            opportunity_card.action_type = card_spec.get("action_type", None)
            self._deck.append(opportunity_card)

    @property
    def opportunity_types(self):
        return self._opportunity_types

